/*
Driver: accesisa
Description: ACCES I/O PC104 Series Driver
Devices:
Author: James Damon
Updated: Tue, 13 May 2014 12:26:55 -0700
Status: experimental

This driver is a documented example on how Comedi drivers are
written.

Configuration Options:
  none
*/

#include <linux/comedidev.h>
/* #include <linux/pci.h> */
/* Imaginary registers for the imaginary board */

#define ACCESISA_SIZE 0

#define ACCESISA_START_AI_CONV	0
#define ACCESISA_AI_READ		0

#define ACCES_READ 0xff00
#define ACCES_WRITE 0x00ff

#define N_CHANS 8


#define MAKE_ADDRESS(X,Y)  ((X << 16 ) && (Y))
#define BASE_ADDRESS(X)    ((0xffff) && (X) )

#define IRQ 1

/*
 * Board descriptions for two imaginary boards.  Describing the
 * boards in this way is optional, and completely driver-dependent.
 * Some drivers use arrays such as this, other do not.
 */
typedef struct accesisa_board_struct {
    const char *name;
    int ai_chans;
    int ai_bits;
    int have_dio;
    /* IRQ values */
    int clear_interupt_offset;
    int clear_interrupt_value;
    int enable_irq_offset;
    int disable_irq_offset;

    int num_dio_bits;
    int num_out_bits;

    int const *read_bases;      /* DI Reading */
    int num_read_bases;

    int const *do_bases;        /* DO Reading */
    int num_do_bases;

    int num_in_bits;
    int const *write_bases;
    int interrupt_status_offset;
    int num;
    int iosize;
} accesisa_board;

static const comedi_lrange waveform_ai_ranges = {
	2,
	{
			BIP_RANGE(10),
			BIP_RANGE(5),
		}
};

static const accesisa_board accesisa_boards[] = {
	{
	      name:	"accesio_idio16",
              iosize:8,
	      ai_chans:0,
	      ai_bits:0,
	      num_dio_bits:0,
              num_out_bits:16,
              clear_interupt_offset:MAKE_ADDRESS(ACCES_WRITE, 0x01 ),
              clear_interrupt_value:1,
              disable_irq_offset:MAKE_ADDRESS(ACCES_WRITE, 0x02 ),
              enable_irq_offset:MAKE_ADDRESS(ACCES_READ, 0x02 ),
              read_bases: (const int []){ MAKE_ADDRESS( ACCES_READ , 0x01), MAKE_ADDRESS(ACCES_READ , 0x05 ) },
              num_read_bases: 2,
              num_in_bits:16,
              write_bases:(const int []){ MAKE_ADDRESS( ACCES_WRITE, 0x01), MAKE_ADDRESS(ACCES_WRITE, 0x04 ) },
        },
	{
	      name:	"accesio_iiro16",
              iosize:8,
	      ai_chans:0,
	      ai_bits:0,
	      num_in_bits:16,
	      num_out_bits:16,
              clear_interupt_offset:MAKE_ADDRESS(ACCES_WRITE, 0x01 ),
              clear_interrupt_value:1,
              disable_irq_offset:MAKE_ADDRESS(ACCES_WRITE, 0x02 ),
              enable_irq_offset: MAKE_ADDRESS(ACCES_READ, 0x02 ),
              read_bases: (const int []){ MAKE_ADDRESS(ACCES_READ  , 0x00 ), /* First relay output */
                                          MAKE_ADDRESS(ACCES_READ  , 0x04 ), /* Second relay output */
                                          MAKE_ADDRESS(ACCES_READ  , 0x01 ), /* First Isolated input */
                                          MAKE_ADDRESS(ACCES_READ  , 0x05 )  /* Second Isolated input */
                                        },
              num_read_bases:4,
              write_bases:(const int []){ MAKE_ADDRESS( ACCES_WRITE, 0x00), MAKE_ADDRESS(ACCES_WRITE, 0x04 ) },
              
        },
        /* Must support 
         * counter
         * AO
         * AI
         * DO
         * DI
         */
	/* { */
	/*       name:	"accesio_aio16", */
        /*       iosize: 0x1E, */
	/*       ai_chans:0, */
	/*       ai_bits:16, */
	/*       num_dio_bits:16, */
        /*       clear_interupt_offset:MAKE_ADDRESS(ACCES_WRITE, 0x01 ), */
        /*       clear_interrupt_value:1, */
        /*       disable_irq_offset:MAKE_ADDRESS(ACCES_WRITE, 0x02 ), */
        /*       enable_irq_offset: MAKE_ADDRESS(ACCES_READ, 0x02 ), */
        /*       read_bases: (const int []){ MAKE_ADDRESS( ACCES_READ , 0x01), MAKE_ADDRESS(ACCES_READ , 0x05 ) }, */
        /*       write_bases:(const int []){ MAKE_ADDRESS( ACCES_WRITE, 0x00), MAKE_ADDRESS(ACCES_WRITE, 0x04 ) }, */
        /* } */
};

/* Comedi Constants and Ranges*/
static const comedi_lrange range_accesisa_aio_range = { 2, { BIP_RANGE(10) , UNI_RANGE(5)  }
};

/* This is used by modprobe to translate PCI IDs to drivers.  Should
 * only be used for PCI and ISA-PnP devices */
/* Please add your PCI vendor ID to comedidev.h, and it will be forwarded
 * upstream. */

/*
 * Useful for shorthand access to the particular board structure
 */
#define thisboard ((const accesisa_board *)dev->board_ptr)

/* this structure is for data unique to this hardware driver.  If
   several hardware drivers keep similar information in this structure,
   feel free to suggest moving the variable to the comedi_device struct.  */
typedef struct {
    int data;
    int base;
    /* would be useful for a PCI device */
    /* struct pci_dev *pci_dev; */
    /* Used for AO readback */
    lsampl_t ao_readback[2];
    volatile lsampl_t ao_loopbacks[N_CHANS];

    struct timer_list timer;
    struct timeval last;	// time at which last timer interrupt occured
    unsigned int uvolt_amplitude;	// waveform amplitude in microvolts
    unsigned long usec_period;	// waveform period in microseconds
    volatile unsigned long usec_current;	// current time (modulo waveform period)
    volatile unsigned long usec_remainder;	// usec since last scan;
    volatile unsigned long ai_count;	// number of conversions remaining
    unsigned int scan_period;	// scan period in usec
    unsigned int convert_period;	// conversion period in usec
    volatile unsigned timer_running:1;


} accesisa_private;

/*
 * most drivers define the following macro to make it easy to
 * access the private structure.
 */
#define devpriv ((accesisa_private *)dev->private)

/*
 * The comedi_driver structure tells the Comedi core module
 * which functions to call to configure/deconfigure (attach/detach)
 * the board, and also about the kernel module that contains
 * the device code.
 */
static int accesisa_attach(comedi_device * dev, comedi_devconfig * it);
static int accesisa_detach(comedi_device * dev);
static comedi_driver driver_accesisa = {
      driver_name:"dummy",
      module:THIS_MODULE,
      attach:accesisa_attach,
      detach:accesisa_detach,
      board_name:&accesisa_boards[0].name,
      offset:sizeof(accesisa_board),
      num_names:sizeof(accesisa_boards) / sizeof(accesisa_board),
};

/* static int waveform_ai_cmdtest(comedi_device * dev, comedi_subdevice * s, */
/*                                comedi_cmd * cmd); */
static int waveform_ai_cmd(comedi_device * dev, comedi_subdevice * s);
static int waveform_ai_cancel(comedi_device * dev, comedi_subdevice * s);

static int waveform_ai_insn_read(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);

#if 0
static int accesisa_cmd_irq(comedi_device * dev, comedi_subdevice * s);

static int accesisa_ai_rinsn(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int accesisa_ao_winsn(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int accesisa_ao_rinsn(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
#endif

static int accesisa_ai_cmdtest(comedi_device * dev, comedi_subdevice * s,
	comedi_cmd * cmd);

static int accesisa_dio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int accesisa_dio_insn_config(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);

static const int nano_per_micro = 1000;	// 1000 nanosec in a microsec


static int start_interrupt( comedi_device *dev );

static lsampl_t tmpval = 0;

static int mytest_open(comedi_device * dev)
{
    int result;
    printk("comedi%d: Opening device !!!\n", dev->minor);
    tmpval = 0;
    result = 0;
    return result;
}

static void mytest_close(comedi_device * dev)
{
    int result;
    printk("comedi%d: Closing device !!!\n", dev->minor);
    result = 0;
}

int read_card( comedi_device * dev )
{
        int tmp = 0;
        printk("comedi%d: performing inb( 0x%x )\n",dev->minor , (unsigned)dev->iobase + 1  );
        tmp =  (0xFF & ~inb( dev->iobase + 0x1 ));
        printk("comedi%d: performing inb( 0x%x )\n",dev->minor , (unsigned)dev->iobase + 5 );
        tmp |= ( (0xFF & (~inb( dev->iobase + 0x5 ))) << 8 ); 
        return tmp;
}

static irqreturn_t accesisa_irq_handler(int irq, void *d PT_REGS_ARG)
{
  comedi_device *dev  = d;
  unsigned long flags;
  sampl_t tmp;
  comedi_subdevice *s = dev->read_subdev;
  printk("comedi%d: Handled interrupt: %d !\n", dev->minor, irq );
  comedi_spin_lock_irqsave(&dev->spinlock, flags);
  
  /* First get some data
     is our issue that lsampl_t is 32bit and we are only returning
     16 bit
  */
  
  tmp = (sampl_t)read_card( dev );
  /* /\* Believe it requires having a buffer first *\/ */
  /* printk("comedi%d: was going to write to buffer\n", dev->minor ); */
   /* cfc_write_long_to_buffer( dev->read_subdev, tmp ); */
  /* printk("comedi%d: Size is %d\n", dev->minor, dev->read_subdev->async->prealloc_bufsz ); */
  /* Clear interrupt */
  outb(0xF, dev->iobase + 0x1 );

  comedi_spin_unlock_irqrestore(&dev->spinlock, flags);

  printk("comedi%d: Sending event to subdevice %d\n", dev->minor, 2 );
  printk("Addres of subdev: %p\n", s );
  /* cfc_write_to_buffer( dev->read_subdev, tmp ); */
  /* s->async->events |= COMEDI_CB_EOS | COMEDI_CB_BLOCK ; /\* Indicate an end of scan *\/ */
  comedi_buf_put(s->async, tmp );
 
  /* Even this caused a failure */
  s->async->events |= COMEDI_CB_BLOCK | COMEDI_CB_EOS;
  /* printk("s->async: %p\n", s->async ); */
  /* s->async->events |= COMEDI_CB_EOS; /\* End of scan signal *\/ */
  comedi_event(dev, s);

  return IRQ_HANDLED;
}

/**
 * @todo Loop over the various types of boards, and set the appropriate
 *       subdevices accordingly.
 */
static int accesisa_attach(comedi_device * dev, comedi_devconfig * it)
{

	comedi_subdevice *s;
        int ret;
        /* Assign base address */
	printk("comedi%d: accesisa: ", dev->minor);

        /* Assign base address */
        
	dev->board_name = thisboard->name;
	printk("tried to assign name '%s' ", dev->board_name );

	if (alloc_private(dev, sizeof(accesisa_private)) < 0)
		return -ENOMEM;

        if ( it->options[0] < 0 ) { 
                return -ENOENT;
        }
        dev->iobase = it->options[0];
        if ( dev->iobase < 0 ) {
                printk("valid io base address required\n");
                return -EINVAL;
        }
        dev->irq = it->options[1];
        printk(" using base=0x%x ", (unsigned)dev->iobase );

        if ( !request_region( dev->iobase, thisboard->iosize, thisboard->name ) ) {
            printk("I/O port conflict: unable to allocate range 0x%lx to 0x%lx\n", dev->iobase, dev->iobase + thisboard->iosize-1 );
            return -EIO;
        }

        printk(" %s " , "Setting up Irq handler ");
        if (dev->irq == 0) {
          printk(" unknown irq (bad)\n");
        } else {
          if ((ret = comedi_request_irq(dev->irq, 
                                        accesisa_irq_handler,
                                        IRQF_SHARED, 
                                        "accesisa", 
                                        dev)) < 0) {
            printk(" Irq %d not available\n", dev->irq );
            dev->irq = 0;
          } else {
            printk(" Irq setup !!\n");
          }
        }

        dev->open = mytest_open;
        dev->close = mytest_close;
        /*
         * Allocate the subdevice structures.  alloc_subdevice() is a
         * convenient macro defined in comedidev.h.
         */
	if (alloc_subdevices(dev, 3) < 0)
		return -ENOMEM;

        /*------------------------------  FIRST DEVICE -------------------------------*/
        /* Interupt */
	s = dev->subdevices + 0;
	dev->read_subdev = s;
        s->type = COMEDI_SUBD_DI;
	s->subdev_flags = SDF_READABLE | SDF_CMD_READ | SDF_GROUND ;
	/* s->n_chan = thisboard->num_out_bits; */
        s->n_chan = 8;
	s->len_chanlist = s->n_chan * 2; /* Caused bug */
        s->maxdata = 1;
	s->range_table = &range_digital;
        s->insn_bits    = accesisa_dio_insn_bits;
        s->insn_config  = accesisa_dio_insn_config;
	s->do_cmd       = waveform_ai_cmd;
	s->do_cmdtest   = accesisa_ai_cmdtest;
        /* s->do_cmd       = accesisa_cmd_irq; */

        /*------------------------------  NEXT DEVICE  -------------------------------*/
	s = dev->subdevices + 1;
	dev->read_subdev = s;
        s->type = COMEDI_SUBD_DI;
	s->subdev_flags = SDF_READABLE | SDF_CMD_READ | SDF_GROUND ;
	s->len_chanlist = s->n_chan * 2;
        s->maxdata = 1;
	s->range_table = &range_digital;
        s->insn_bits    = accesisa_dio_insn_bits;
        s->insn_config  = accesisa_dio_insn_config;
	s->do_cmd       = waveform_ai_cmd;
	s->do_cmdtest   = accesisa_ai_cmdtest;
        s->n_chan = thisboard->num_in_bits;


	/* Digital Output subdevice */

        /*-----------------------------  THIRD DEVICE  -----------------------------*/
	s = dev->subdevices + 2;
        s->type = COMEDI_SUBD_AI;
        s->subdev_flags = SDF_READABLE | SDF_GROUND | SDF_CMD_READ;
        s->n_chan = thisboard->ai_chans;
        s->maxdata = (1 << thisboard->ai_bits) - 1;
        s->range_table = &waveform_ai_ranges;
        s->len_chanlist = s->n_chan * 2;
        s->insn_read = waveform_ai_insn_read;
        s->do_cmd = waveform_ai_cmd;
        s->do_cmdtest = accesisa_ai_cmdtest;
        s->cancel = waveform_ai_cancel;


	printk("attached\n");

	return 0;
}

static int start_interrupt( comedi_device *dev )
{
  printk("Starting the interrupt function\n");
  /* Disable IRQ */
  /* outb(0xF, dev->iobase + 0x2 ); */
#if 1
  outb(0xF, 0x302 );
  /* Clear IRQ */
  /* outb(0xF, dev->iobase + 0x1 ); */
  outb(0xF, 0x301 );
  /* Enable IRQ */
  /* inb(dev->iobase + 0x2 ); */
  inb(0x302 );
#endif
  /* s->async->events |= COMEDI_CB_EOS COMEDI_CB_BLOCK ; /\* Indicate an end of scan *\/ */
  return 0;
}

/*
 * _detach is called to deconfigure a device.  It should deallocate
 * resources.
 * This function is also called when _attach() fails, so it should be
 * careful not to release resources that were not necessarily
 * allocated by _attach().  dev->private and dev->subdevices are
 * deallocated automatically by the core.
 */
static int accesisa_detach(comedi_device * dev)
{
	printk("comedi%d: accesisa: remove\n", dev->minor);
        if ( dev->iobase ) {
            release_region( dev->iobase, thisboard->iosize );
        }

        printk("comedi%d: HERE!!: IRQ:%d\n", dev->minor, dev->irq );
        if (dev->irq ) {
          printk("comedi%d: removing the irq %d\n", dev->minor, dev->irq );
          comedi_free_irq(dev->irq, dev);
        }

	return 0;
}

#if 0
static int accesisa_ai_rinsn(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
        int n = 0;
	/* return the number of samples read/written */
	return n;
}
#endif

static int waveform_ai_insn_read(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
	int i, chan = CR_CHAN(insn->chanspec);

	for (i = 0; i < insn->n; i++)
		data[i] = devpriv->ao_loopbacks[chan];

	return insn->n;
}

#if 0

static int accesisa_cmd_irq(comedi_device * dev, comedi_subdevice * s)
{
  printk("comedi%d: Setting up irq\n", dev->minor );
  comedi_cmd *cmd = &s->async->cmd;
  /* unsigned long flags; */
  /* comedi_spin_lock_irqsave(&subpriv->spinlock, flags); */
  /* subpriv->active = 1; */
  /* Set up end of acquisition. */
  /* Set up start of acquisition. */
#if 0
  switch (cmd->start_src) {
  case TRIG_INT:
    /* s->async->inttrig = dio200_inttrig_start_intr; /\*  *\/ */
    s->async->inttrig = start_interrupt;
    break;
  default:
    /* TRIG_NOW */
    /* event = dio200_start_intr(dev, s); */
    event = start_interrupt( dev );
    break;
  }
  if (event) {
    comedi_event(dev, s);
  }

#endif

  return 0;
}
#endif

static int waveform_ai_cmd(comedi_device * dev, comedi_subdevice * s)
{
	comedi_cmd *cmd = &s->async->cmd;
        int event;
	if (cmd->flags & TRIG_RT) {
		comedi_error(dev,
			"commands at RT priority not supported in this driver");
		return -1;
	}

        if ( cmd->start_src == TRIG_INT ) {
          printk("Doing some TRIG_INT\n");
        } else {
          printk("Starting the interrupt !\n");
          event = start_interrupt(dev );
          comedi_event( dev, s);
        }

	if (cmd->convert_src == TRIG_NOW)
		/* devpriv->convert_period = 0; */
          printk("comedi%d: Doing some TRIG_NOW stuff\n", dev->minor);
	else if (cmd->convert_src == TRIG_TIMER)
		/* devpriv->convert_period = cmd->convert_arg / nano_per_micro; */
          printk("comedi%d: Doing some TRIG_TIMER stuff\n",dev->minor);
	else {
          comedi_error(dev, "bug setting conversion period");
          return -1;
	}

	return 0;
}
#define MYDEBUG(str,counter,err) printk(str,dev->minor,(char)((counter++)+65),err);


/* static int waveform_ai_cmdtest(comedi_device * dev, comedi_subdevice * s, */
/* 	comedi_cmd * cmd) */
/* { */
/* 	int err = 0; */
/* 	int tmp; */
/*         int counter = 0; */
/* 	/\* step 1: make sure trigger sources are trivially valid *\/ */
/*         printk("comedi%d: Calling cmd_test\n",dev->minor); */
/*         printk("Start src is %d\n", cmd->start_src ); */
/* 	tmp = cmd->start_src; */
/* 	cmd->start_src &= (TRIG_NOW | TRIG_INT); */
/* 	if (!cmd->start_src || tmp != cmd->start_src) */
/* 		err++; */
/*         MYDEBUG("comedi%d: %c %d\n",counter,err); */

/* 	tmp = cmd->scan_begin_src; */
/* 	/\* cmd->scan_begin_src &= ( TRIG_TIMER | TRIG_FOLLOW ); *\/ */
/* 	cmd->scan_begin_src &= ( TRIG_TIMER | TRIG_FOLLOW ); */
/* 	if (!cmd->scan_begin_src || tmp != cmd->scan_begin_src) */
/* 		err++; */
/*         MYDEBUG("comedi%d: %c %d\n",counter,err); */

/* 	tmp = cmd->convert_src; */
/* 	/\* cmd->convert_src &= TRIG_NOW | TRIG_TIMER; *\/ */
/* 	cmd->convert_src &= TRIG_NOW; */
/* 	if (!cmd->convert_src || tmp != cmd->convert_src) */
/* 		err++; */
/*         MYDEBUG("comedi%d: %c %d\n",counter,err); */

/* 	tmp = cmd->scan_end_src; */
/* 	cmd->scan_end_src &= TRIG_COUNT; */
/*         printk("comedi_test: scan_end_src: %d\n", cmd->scan_end_src ); */
/* 	if (!cmd->scan_end_src || tmp != cmd->scan_end_src) */
/* 		err++; */

/* 	tmp = cmd->stop_src; */
/*         printk("comedi_test: stop_src: %d\n", cmd->stop_src ); */
/* 	cmd->stop_src &= TRIG_NONE; */
/* 	if (!cmd->stop_src || tmp != cmd->stop_src) */
/* 		err++; */

/* 	if (err) */
/* 		return 1; */

/* 	/\* step 2: make sure trigger sources are unique and mutually compatible *\/ */

/* 	if (cmd->convert_src != TRIG_NOW && cmd->convert_src != TRIG_TIMER) */
/* 		err++; */
/* 	if (cmd->stop_src != TRIG_COUNT && cmd->stop_src != TRIG_NONE) */
/* 		err++; */

/* 	if (err) */
/* 		return 2; */

/* 	/\* step 3: make sure arguments are trivially compatible *\/ */

/* 	if (cmd->start_arg != 0) { */
/* 		cmd->start_arg = 0; */
/* 		err++; */
/* 	} */

/* 	if (cmd->convert_src == TRIG_NOW) { */
/* 		if (cmd->convert_arg != 0) { */
/* 			cmd->convert_arg = 0; */
/* 			err++; */
/* 		} */
/* 	} */
/* 	if (cmd->scan_begin_src == TRIG_TIMER) { */
/* 		if (cmd->scan_begin_arg < nano_per_micro) { */
/* 			cmd->scan_begin_arg = nano_per_micro; */
/* 			err++; */
/* 		} */
/* 		if (cmd->convert_src == TRIG_TIMER && */
/* 			cmd->scan_begin_arg < */
/* 			cmd->convert_arg * cmd->chanlist_len) { */
/* 			cmd->scan_begin_arg = */
/* 				cmd->convert_arg * cmd->chanlist_len; */
/* 			err++; */
/* 		} */
/* 	} */
/* 	// XXX these checks are generic and should go in core if not there already */
/* 	if (!cmd->chanlist_len) { */
/* 		cmd->chanlist_len = 1; */
/* 		err++; */
/* 	} */
/* 	if (cmd->scan_end_arg != cmd->chanlist_len) { */
/* 		cmd->scan_end_arg = cmd->chanlist_len; */
/* 		err++; */
/* 	} */

/* 	if (cmd->stop_src == TRIG_COUNT) { */
/* 		if (!cmd->stop_arg) { */
/* 			cmd->stop_arg = 1; */
/* 			err++; */
/* 		} */
/* 	} else {		/\* TRIG_NONE *\/ */
/* 		if (cmd->stop_arg != 0) { */
/* 			cmd->stop_arg = 0; */
/* 			err++; */
/* 		} */
/* 	} */

/* 	if (err) */
/* 		return 3; */

/* 	/\* step 4: fix up any arguments *\/ */

/* 	if (cmd->scan_begin_src == TRIG_TIMER) { */
/* 		tmp = cmd->scan_begin_arg; */
/* 		// round to nearest microsec */
/* 		cmd->scan_begin_arg = */
/* 			nano_per_micro * ((tmp + */
/* 				(nano_per_micro / 2)) / nano_per_micro); */
/* 		if (tmp != cmd->scan_begin_arg) */
/* 			err++; */
/* 	} */
/* 	if (cmd->convert_src == TRIG_TIMER) { */
/* 		tmp = cmd->convert_arg; */
/* 		// round to nearest microsec */
/* 		cmd->convert_arg = */
/* 			nano_per_micro * ((tmp + */
/* 				(nano_per_micro / 2)) / nano_per_micro); */
/* 		if (tmp != cmd->convert_arg) */
/* 			err++; */
/* 	} */

/* 	if (err) */
/* 		return 4; */

/* 	return 0; */
/* } */

static int accesisa_ai_cmdtest(comedi_device * dev, comedi_subdevice * s,
	comedi_cmd * cmd)
{
	int err = 0;
	int tmp;
        int counter = 0;
	/* step 1: make sure trigger sources are trivially valid */
        printk("comedi%d: Calling cmd_test\n",dev->minor);
        printk("Start src is %d\n", cmd->start_src );
	tmp = cmd->start_src;
	cmd->start_src &= (TRIG_NOW | TRIG_INT);
	if (!cmd->start_src || tmp != cmd->start_src)
		err++;
        MYDEBUG("comedi%d: %c %d\n",counter,err);

	tmp = cmd->scan_begin_src;
	/* cmd->scan_begin_src &= ( TRIG_TIMER | TRIG_FOLLOW ); */
	cmd->scan_begin_src &= ( TRIG_TIMER | TRIG_FOLLOW );
	if (!cmd->scan_begin_src || tmp != cmd->scan_begin_src)
		err++;
        MYDEBUG("comedi%d: %c %d\n",counter,err);

	tmp = cmd->convert_src;
	/* cmd->convert_src &= TRIG_NOW | TRIG_TIMER; */
	cmd->convert_src &= TRIG_NOW;
	if (!cmd->convert_src || tmp != cmd->convert_src)
		err++;
        MYDEBUG("comedi%d: %c %d\n",counter,err);

	tmp = cmd->scan_end_src;
	cmd->scan_end_src &= TRIG_COUNT;
        printk("comedi_test: scan_end_src: %d\n", cmd->scan_end_src );
	if (!cmd->scan_end_src || tmp != cmd->scan_end_src)
		err++;

	tmp = cmd->stop_src;
        printk("comedi_test: stop_src: %d\n", cmd->stop_src );
	cmd->stop_src &= TRIG_NONE;
	if (!cmd->stop_src || tmp != cmd->stop_src)
		err++;

	if (err)
		return 1;

	/* step 2: make sure trigger sources are unique and mutually compatible */

	if (cmd->convert_src != TRIG_NOW && cmd->convert_src != TRIG_TIMER)
		err++;
	if (cmd->stop_src != TRIG_COUNT && cmd->stop_src != TRIG_NONE)
		err++;

	if (err)
		return 2;

	/* step 3: make sure arguments are trivially compatible */

	if (cmd->start_arg != 0) {
		cmd->start_arg = 0;
		err++;
	}

	if (cmd->convert_src == TRIG_NOW) {
		if (cmd->convert_arg != 0) {
			cmd->convert_arg = 0;
			err++;
		}
	}
	if (cmd->scan_begin_src == TRIG_TIMER) {
		if (cmd->scan_begin_arg < nano_per_micro) {
			cmd->scan_begin_arg = nano_per_micro;
			err++;
		}
		if (cmd->convert_src == TRIG_TIMER &&
			cmd->scan_begin_arg <
			cmd->convert_arg * cmd->chanlist_len) {
			cmd->scan_begin_arg =
				cmd->convert_arg * cmd->chanlist_len;
			err++;
		}
	}
	// XXX these checks are generic and should go in core if not there already
	if (!cmd->chanlist_len) {
		cmd->chanlist_len = 1;
		err++;
	}
	if (cmd->scan_end_arg != cmd->chanlist_len) {
		cmd->scan_end_arg = cmd->chanlist_len;
		err++;
	}

	if (cmd->stop_src == TRIG_COUNT) {
		if (!cmd->stop_arg) {
			cmd->stop_arg = 1;
			err++;
		}
	} else {		/* TRIG_NONE */
		if (cmd->stop_arg != 0) {
			cmd->stop_arg = 0;
			err++;
		}
	}

	if (err)
		return 3;

	/* step 4: fix up any arguments */

	if (cmd->scan_begin_src == TRIG_TIMER) {
		tmp = cmd->scan_begin_arg;
		// round to nearest microsec
		cmd->scan_begin_arg =
			nano_per_micro * ((tmp +
				(nano_per_micro / 2)) / nano_per_micro);
		if (tmp != cmd->scan_begin_arg)
			err++;
	}
	if (cmd->convert_src == TRIG_TIMER) {
		tmp = cmd->convert_arg;
		// round to nearest microsec
		cmd->convert_arg =
			nano_per_micro * ((tmp +
				(nano_per_micro / 2)) / nano_per_micro);
		if (tmp != cmd->convert_arg)
			err++;
	}

	if (err)
		return 4;

	return 0;

}



#if 0
static int accesisa_ao_winsn(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
	int i;
	int chan = CR_CHAN(insn->chanspec);

	printk("accesisa_ao_winsn\n");

	return i;
}

/* AO subdevices should have a read insn as well as a write insn.
 * Usually this means copying a value stored in devpriv. */
static int 
accesisa_ao_rinsn(comedi_device * dev, comedi_subdevice * s,
                  comedi_insn * insn, lsampl_t * data)
{
	int i;
	int chan = CR_CHAN(insn->chanspec);

	for (i = 0; i < insn->n; i++)
		data[i] = devpriv->ao_readback[chan];
        printk("comedi%d doing something in the accesisa_ao_rinsn\n", dev->minor );
        /* inb( BASE_ADDRESS ); */

	return i;
}
#endif


/**
 * @desc 
 *
 */
static int 
accesisa_dio_insn_config(comedi_device * dev, comedi_subdevice * s,
                         comedi_insn * insn, lsampl_t * data)
{
	int chan = CR_CHAN(insn->chanspec);
        printk("comedi%d: DIO configuring\n", dev->minor);
	/* The input or output configuration of each digital line is
	 * configured by a special insn_config instruction.  chanspec
	 * contains the channel to be changed, and data[0] contains the
	 * value COMEDI_INPUT or COMEDI_OUTPUT. */
	switch (data[0]) {
	case INSN_CONFIG_DIO_OUTPUT:
            printk("comedi%d: DIO Output\n", dev->minor);
            s->io_bits |= 1 << chan;
            break;
	case INSN_CONFIG_DIO_INPUT:
            printk("comedi%d: DIO Input\n", dev->minor);
            s->io_bits &= ~(1 << chan);
            break;
	case INSN_CONFIG_DIO_QUERY:
            printk("comedi%d: DIO Query\n", dev->minor);
            data[1] = ( s->io_bits & (1 << chan)) ? COMEDI_OUTPUT : COMEDI_INPUT;
            return insn->n;
            break;
	default:
		return -EINVAL;
		break;
	}

	return insn->n;
}


char *lookup_insn( unsigned int insn ) {
    switch(insn) {
    case INSN_READ:
        return "INSN_READ";
    case INSN_WRITE:
        return "INSN_WRITE";
    case INSN_BITS:
        return "INSN_BITS";
    case INSN_CONFIG:
        return "INSN_CONFIG";
    case INSN_GTOD:
        return "INSN_GTOD";
    case INSN_WAIT:
        return "INSN_WAIT";
    default:
        return "";
    }
}

void debug_comedi_insn( comedi_insn *insn )
{
    printk("INSN:\n");
    printk("  insn: %s\n", lookup_insn(insn->insn) );
    printk("  num: %d\n", insn->n );
}

static int waveform_ai_cancel(comedi_device * dev, comedi_subdevice * s)
{
	devpriv->timer_running = 0;
	del_timer_sync(&devpriv->timer);
	return 0;
}


/* DIO devices are slightly special.  Although it is possible to
 * implement the insn_read/insn_write interface, it is much more
 * useful to applications if you implement the insn_bits interface.
 * This allows packed reading/writing of the DIO channels.  The
 * comedi core can convert between insn_bits and insn_read/write */
static int accesisa_dio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
        /* int i; */
        debug_comedi_insn( insn );

        if (  insn->insn == INSN_READ ) {
            printk(	"comedi%d: Reading something\n", dev->minor);
        } else if ( insn->insn == INSN_WRITE ) {
            printk(	"comedi%d: Writing something\n", dev->minor);            
        } else if ( insn->insn == INSN_BITS ) {
            sampl_t temp;
            /* data[1] = tmpval; */
            printk(	"comedi%d: Bits something: %d, data=%d\n", dev->minor, (int)tmpval, (int)data[1] );            
            /**
             * @note Can i do a comedi_buf_get( ) here to grab
             *       data from the buffer....or instead, should I just 
             *       do the straight inb from the appropriate base address...
             */
            comedi_buf_get(s->async, &temp );
            data[0] = temp;
            /* ret = comedi_buf_get(s->async, &temp); */
            /* data[0] = temp; */
            /* data[1] = temp >> 8; */
        }

        printk("comedi%d: data[0]=%d\n", dev->minor, data[0] );
        printk("comedi%d: data[1]=%d\n", dev->minor, data[1] );
        /* Debug the instruction */
        if (insn->n != 2)       /* What does this actually mean here... */
		return -EINVAL;

        printk(	"comedi%d: Here\n", dev->minor );
        data[0] = 0;
        /* for ( i = 0 ; i < thisboard->num_read_bases ; i ++ ) { */
        /*     printk("comedi%d: performing inb( 0x%x )\n", dev->minor, dev->iobase + thisboard->read_bases[i] ); */
        /*     data[0] = tmpval; */
        /* } */

        /* Checking something...is this necessary */
        /* You need to replace this part ... */
        if (data[0]) {
                /* Check if requested ports are configured for output */
                /* if ((s->io_bits & data[0]) != data[0]) */
                /*         return -EIO; */
                /* s->state &= ~data[0]; */
                s->state |= data[0] & data[1];
        }

	return 2;
}

/* The insn data is a mask in data[0] and the new data
 * in data[1], each channel cooresponding to a bit. */
/* data[0] = inb( dev->iobase + correct_offset ); */
/* data[1] = inb( dev->iobase + correct_second_offset ); */
/* on return, data[1] contains the value of the digital
 * input and output lines. */
//data[1]=inw(dev->iobase + ACCESISA_DIO);
/* or we could just return the software copy of the output values if
 * it was a purely digital output subdevice */
//data[1]=s->state;


/*
 * A convenient macro that defines init_module() and cleanup_module(),
 * as necessary.
 */
COMEDI_INITCLEANUP(driver_accesisa);
/* If you are writing a PCI driver you should use COMEDI_PCI_INITCLEANUP instead.
*/
// COMEDI_PCI_INITCLEANUP(driver_accesisa, accesisa_pci_table)

/* GARBAGE */

/* /\** */
/*  * @desc  */
/*  * */
/*  *\/ */
/* static int skel_dio_insn_config(comedi_device * dev, comedi_subdevice * s, */
/* 	comedi_insn * insn, lsampl_t * data) */
/* { */
/* 	int chan = CR_CHAN(insn->chanspec); */
/* 	/\* The input or output configuration of each digital line is */
/* 	 * configured by a special insn_config instruction.  chanspec */
/* 	 * contains the channel to be changed, and data[0] contains the */
/* 	 * value COMEDI_INPUT or COMEDI_OUTPUT. *\/ */
/* 	switch (data[0]) { */
/* 	case INSN_CONFIG_DIO_OUTPUT: */
/* 		s->io_bits |= 1 << chan; */
/* 		break; */
/* 	case INSN_CONFIG_DIO_INPUT: */
/* 		s->io_bits &= ~(1 << chan); */
/* 		break; */
/* 	case INSN_CONFIG_DIO_QUERY: */
/* 		data[1] = ( s->io_bits & (1 << chan)) ? COMEDI_OUTPUT : COMEDI_INPUT; */
/* 		return insn->n; */
/* 		break; */
/* 	default: */
/* 		return -EINVAL; */
/* 		break; */
/* 	} */
/* 	//outw(s->io_bits,dev->iobase + SKEL_DIO_CONFIG); */
/* 	return insn->n; */
/* } */
