/*
   comedi/drivers/accespci.c
   Source code for the ACCES I/O Products PCI board family.

   COMEDI - Linux Control and Measurement Device Interface
   Copyright (C) 2000 David A. Schleef <ds@schleef.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
Driver: accespci
Description: ACCES I/O Products PCI and PCIe series boards
Devices: ACCES I/O Products pci_dio_24, pci_dio_48, pci_iiro_16, pci_iiro_8, pcie_iiro_8
Author: Jimi Damon  <jdamon@accesio.com>)
Updated: Tue, 05 May 2015 15:34:01 -0800
Status: workging (no support for loading firmware)
comedi/drivers/accespci.c

Configuration Options:

    [0] - PCI bus number (optional)
    [1] - PCI slot number (optional)

    If bus/slot is not specified, the first available PCI
    device will be used.
*/

#include <linux/comedidev.h>
#include "apci_common.h"
#include "comedi_pci.h"


int comedi_debug = 1;

#define APCI_SIZE 0

#define APCI_START_AI_CONV	0
#define APCI_AI_READ		0

#define INPUT 1
#define OUTPUT 0
#define MODE_SEL_MODE1 1
#define MODE_SEL_MODE0 0

#define MODE_ACTIVE_AND_TRISTATE 1
#define SIZEOF_ADDRESS_SPACE  0x20
 
#define MODE1_HIGH 1
#define MODE1_LOW  0
#define MODE0_HIGH 0
#define MODE0_LOW  0

#define MAKE_BYTE(A,B,C,D,E,F,G,H) ( ( (A&1) << 7 ) | ( (B&1) << 6 ) | ( (C&1) << 5 ) | ( (D&1) << 4 ) | \
                                     ( (E&1) << 3 ) | ( (F&1) << 2 ) | ( (G&1) << 1 ) | ( (H&1) ) )


typedef struct apci_board_struct {
        const char *name;
        int ai_chans;
        int ai_bits;
        int have_dio;
        int dev_id;
        int number_dio_channels;
        int num_groups;
        int num_ports;
        int (*setup_device)( comedi_device *dev );
        int (*teardown_device)( comedi_device *dev );
        int (*initialize)( comedi_device *dev );
        irqreturn_t  (*irqhandler)( int irq, void *d );
        int is_pcie;
} apci_board;

#define MYDEBUG(str,counter,err) printk(str,dev->minor,(char)((counter++)+65),err);

typedef enum {
    PORT_A = 0,
    PORT_OUT = 0,
    PORT_B = 1,
    PORT_IN = 1,
    PORT_C = 2,
    PORT_CHI = 2,
    PORT_CLO = 3
} PORT;

typedef struct {
    int start;
    int length;
    int flags;
    int enabled;
    void *mapped_address;
} io_region_t;


typedef struct {
        struct pci_dev *pcidev;
        int data;
        lsampl_t ao_readback[2];
        unsigned char mode[5];
        void *dio_base;
        int got_regions;
        spinlock_t lock;
        int plx_region_start;
        int plx_region_length;
        io_region_t regions[6];
        int is_pcie;
        int numevents;
        int continuous;
        int previous;
        int previous_set;
} apci_private;

typedef struct {
        int offset;
        PORT port;
        int group;
        int numevents;
        int state;
        int continuous;
} subdev_private;


#define PCI_VENDOR_ID_APCI 0x494F
#define PCI_DIO_24 0x0c50
#define PCI_DIO_48 0x0c60
#define PCI_IIRO_8 0x0F00
#define PCIE_IIRO_8 0x0F02
#define PCI_IIRO_16 0x0F08
#define PCIE_IDIO_24 0x0FD0


int setup_device_dio48( comedi_device *dev );
int setup_device_iiro8( comedi_device *dev );

/* IDIO */
int setup_device_idio( comedi_device *dev );
static int teardown_device_idio( comedi_device *dev );
irqreturn_t idio_irqhandler( int irq, void *d PT_REGS_ARG );
static int apci_idio_initialize( comedi_device *dev );

static int apci_idio_fetout_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int apci_idio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int apci_idio_insn_ttl_config( comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int apci_idio_ttl_cmdtest( comedi_device *dev, comedi_subdevice *s, comedi_cmd *cmd );
static int apci_idio_ttl_cmd( comedi_device * dev, comedi_subdevice *s );
static int apci_idio_cancel( comedi_device * dev, comedi_subdevice *s );
static int teardown_device_iiro( comedi_device *dev );



/* Generic DIO */
int teardown_devices_generic_dio( comedi_device *dev );
irqreturn_t iiro_irqhandler( int irq, void *d PT_REGS_ARG );
static int apci_iiro_intr_cmdtest(comedi_device * dev, comedi_subdevice * s,comedi_cmd * cmd);
static int apci_iiro_intr_cmd(comedi_device * dev, comedi_subdevice * s);
static int apci_iiro_initialize( comedi_device * dev );

static const int nano_per_micro = 1000;	// 1000 nanosec in a microsec


static const apci_board apci_boards[] = {
        {
        name:	"pci_iiro_16",
        dev_id: PCI_IIRO_16,
        ai_chans: 0,
        ai_bits: 0,
        have_dio:1,
        num_groups:2,
        num_ports:2,
        is_pcie:0,
        setup_device:setup_device_iiro8,
        teardown_device:teardown_device_iiro,
        irqhandler:iiro_irqhandler,
        initialize: apci_iiro_initialize,
        },
        {
        name:	"pci_dio_24",
        dev_id:  PCI_DIO_24 ,
        ai_chans:16,
        ai_bits:	12,
        have_dio:1,
        num_groups:1,
        num_ports:3,
        setup_device:setup_device_dio48,
        },
        {
        name:	"pci_dio_48",
        dev_id: PCI_DIO_48,
        ai_chans:8,
        ai_bits:	16,
        have_dio:1,
        num_groups:2,
        num_ports:3,
        is_pcie:0,
        setup_device:setup_device_dio48,
        },
        {
        name:	"pci_iiro_8",
        dev_id: PCI_IIRO_8,
        ai_chans: 0,
        ai_bits: 0,
        have_dio:1,
        num_groups:1,
        num_ports:2,
        is_pcie:0,
        setup_device:setup_device_dio48,
        },
        {
        name:	"pcie_iiro_8",
        dev_id: PCIE_IIRO_8,
        ai_chans: 0,
        ai_bits: 0,
        have_dio:1,
        num_groups:1,
        num_ports:2,
        setup_device:setup_device_iiro8,
        teardown_device:teardown_device_iiro,
        irqhandler: iiro_irqhandler,
        initialize: apci_iiro_initialize,
        is_pcie:1
        },
        {
        name:	"pcie_idio_24",
        dev_id: PCIE_IDIO_24,
        ai_chans: 0,
        ai_bits: 0,
        have_dio:1,
        num_groups:1,
        num_ports:3,
        is_pcie:1,
        setup_device:setup_device_idio,
        teardown_device:teardown_device_idio,
        irqhandler:idio_irqhandler,
        initialize: apci_idio_initialize,
        },
        {
        name: "accespci",
        },
};


static DEFINE_PCI_DEVICE_TABLE(apci_pci_table) = {
        { PCI_DEVICE( PCI_VENDOR_ID_APCI, PCI_DIO_24 ), },
        { PCI_DEVICE( PCI_VENDOR_ID_APCI, PCI_DIO_48 ), },
        { PCI_DEVICE( PCI_VENDOR_ID_APCI, PCI_IIRO_8 ), },
        { PCI_DEVICE( PCI_VENDOR_ID_APCI, PCIE_IIRO_8 ), },
        { PCI_DEVICE( PCI_VENDOR_ID_APCI, PCI_IIRO_16 ), },
        { PCI_DEVICE( PCI_VENDOR_ID_APCI, PCIE_IDIO_24 ), },
        {0,}
};

MODULE_DEVICE_TABLE(pci, apci_pci_table);

#define n_boardtypes (sizeof( apci_boards)/sizeof(apci_board))
#define thisboard ((const apci_board *)dev->board_ptr)
#define devpriv ((apci_private *)dev->private)
#define subdevpriv ((subdev_private *)(s->private))
#define subdevice_private(d,n) ( ((subdev_private*)(&d->subdevices[n]))->private );


static int apci_attach(comedi_device * dev, comedi_devconfig * it);
static int apci_detach(comedi_device * dev);


static comedi_driver driver_apci = {
      driver_name:"accespci",
      module:THIS_MODULE,
      attach:apci_attach,
      detach:apci_detach,
      num_names:n_boardtypes,
      board_name:&apci_boards[0].name,
      offset:sizeof(apci_board)
};

static int apci_dio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int apci_dio_insn_config(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int apci_iiro_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int apci_iiro_insn_config(comedi_device * dev, comedi_subdevice * s,
        comedi_insn * insn, lsampl_t * data);
static int start_interrupt( comedi_device *dev );
static int start_idio_interrupt( comedi_device *dev , comedi_subdevice *s );
static int disable_idio_interrupt( comedi_device *dev , comedi_subdevice *s );

char debug_port( PORT port )
{
    static char retval;
    switch (port) {
    case PORT_A:
        retval = 'A';
        break;
    case PORT_B:
        retval = 'B';
        break;
    case PORT_C:
        retval = 'C';
        break;
    default:
        retval =  '_' ;
        break;
    }
    return retval;
}

char debug_iiro_port( PORT port )
{
    static char retval;
    switch (port) {
    case PORT_OUT:
        retval = 'O';
        break;
    case PORT_IN:
        retval = 'I';
        break;
    default:
        retval =  '_' ;
        break;
    }
    return retval;
}

char *debug_byte( unsigned char byte ) 
{
    static char buf[9];
    for( int i = 7 ; i >= 0 ; i -- ) { 
        /* apci_debug("%d", (int)( byte & (1 << i ))); */
        buf[7-i] = ( byte & ( 1 << i) ? '1' : '0' );
    }
    buf[8] = 0;
    return buf;
}

char *debug_sample( sampl_t sample ) 
{
    static char buf[17];
    for( int i = 15 ; i >= 0 ; i -- ) { 
        /* apci_debug("%d", (int)( byte & (1 << i ))); */
        buf[15-i] = ( sample & ( 1 << i) ? '1' : '0' );
    }
    buf[16] = 0;
    return buf;
}

char *debug_int( int sample ) 
{
    static char buf[33];
    for( int i = 31 ; i >= 0 ; i -- ) { 
        /* apci_debug("%d", (int)( byte & (1 << i ))); */
        buf[31-i] = ( sample & ( 1 << i) ? '1' : '0' );
    }
    buf[32] = 0;
    return buf;
}


void cleanup_io_regions( comedi_device *dev )
{

    for (int i = 0; i < 6; i ++) {
        if (devpriv->regions[i].start == 0) {
            continue; /* invalid region */
        }
        if (devpriv->regions[i].flags & IORESOURCE_IO) {
            release_region(devpriv->regions[i].start, devpriv->regions[i].length);
            apci_debug("Releasing bar=%d,region=%#x,len=%#x\n", i, devpriv->regions[i].start, devpriv->regions[i].length );
        } else {
            iounmap(devpriv->regions[i].mapped_address);
            apci_debug("Unmapping bar=%d,region=%#x,mapped=%lx\n", i, devpriv->regions[i].start, (long int)devpriv->regions[i].mapped_address );
            release_mem_region(devpriv->regions[i].start, devpriv->regions[i].length);
            apci_debug("Releasing mem bar=%d,region=%#x,len=%#x\n", i, devpriv->regions[i].start, devpriv->regions[i].length );
        }
    }
}

static int apci_attach(comedi_device * dev, comedi_devconfig * it)
{
	/* comedi_subdevice *s; */
        int retval = 0;
        struct pci_dev *pcidev;
        int index;
        int plx_bar;
        int ret;


	apci_info("TRYING TO ATTACH comedi%d: \n", dev->minor);

	//dev->board_ptr = apci_probe(dev, it);

	if (alloc_private(dev, sizeof(apci_private)) < 0)
 		return -ENOMEM;

        apci_debug("Initial: Start=%#x, length=%#x\n", devpriv->plx_region_start, devpriv->plx_region_length );

        for (pcidev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, NULL);
             pcidev != NULL;
             pcidev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, pcidev)) {
                // is it not a computer boards card?
                apci_debug("looping: %#x %#.4x\n", pcidev->vendor, pcidev->device );
                if (pcidev->vendor != PCI_VENDOR_ID_APCI)
                        continue;
                // loop through cards supported by this driver
                for (index = 0;
                        index < sizeof apci_boards / sizeof(apci_board);
                        index++) {
                        apci_debug("dev_id=%d ==? pcidev->device=%d\n", apci_boards[index].dev_id, pcidev->device );
                        if (apci_boards[index].dev_id != pcidev->device) {
                                continue;
                        }
                        apci_debug("Found a board at index:%d\n",index );

                        // was a particular bus/slot requested?
                        if (it->options[0] || it->options[1]) {
                                // are we on the wrong bus/slot?
                                apci_debug("pcidev->bus->number=%d, it->options[0]=%d,  PCI_SLOT()=%d, id->options[1]=%d\n",
                                        pcidev->bus->number,
                                        it->options[0],
                                        PCI_SLOT(pcidev->devfn),
                                        it->options[1]
                                        );
                                if ( pcidev->bus->number != it->options[0] || PCI_SLOT(pcidev->devfn) != it->options[1] ) {
                                        continue;
                                }
                                apci_debug("Bus numbers match :) \n" );

                        } else {
                                apci_debug("slot not requested :( \n" );
                        }

                        dev->board_ptr = apci_boards + index;
                        
                        devpriv->pcidev = pcidev;

                        goto found;
                }
        }
        apci_info("No supported ACCES I/O boards found\n");
        return -EIO;
        
 found:

        ret = pci_enable_device( pcidev );
        if ( ret < 0 ) {
            apci_error("Unable to Enable PCI device: %d\n", ret );
                return ret;
        }

        dev->board_name = thisboard->name;

        spin_lock_init(&devpriv->lock);

        apci_info("Found a device: %s\n", thisboard->name );

        devpriv->pcidev = pcidev;

        if (pci_resource_flags(pcidev, 0) & IORESOURCE_IO) {
            plx_bar = 0;
        } else {
            plx_bar = 1;
        }

        apci_debug("Using bar=%d\n",plx_bar );
        devpriv->plx_region_start      = pci_resource_start(pcidev, plx_bar);
        if( ! devpriv->plx_region_start ) {
            apci_error("Invalid bar %d on start ", plx_bar );
            return -ENODEV;
        }
        devpriv->plx_region_length     = pci_resource_len(pcidev, plx_bar );
        apci_debug("Start=%#x, length=%#x\n", devpriv->plx_region_start, devpriv->plx_region_length );        

        devpriv->regions[2].start   = pci_resource_start(pcidev, 2);
        devpriv->regions[2].length  = pci_resource_len(pcidev, 2);
        devpriv->regions[2].flags   = pci_resource_flags(pcidev, 2);

        apci_debug("bar[%d], start=%#x,len=%#x\n", 2,devpriv->regions[2].start, devpriv->regions[2].length );


        devpriv->regions[2].enabled = 1;

        /* Setup private stuff */
        apci_debug("Size of devpriv->mode: %d\n", (int)(sizeof(devpriv->mode)) );
        for ( int i = 0; i < sizeof(devpriv->mode) / sizeof(unsigned char ); i ++ ) {
            devpriv->mode[i] = MAKE_BYTE( MODE_ACTIVE_AND_TRISTATE, MODE0_HIGH, MODE0_LOW, INPUT,INPUT,MODE_SEL_MODE0,INPUT,INPUT);
            apci_debug("mode[%d]: %s\n", i, debug_byte( devpriv->mode[i] ) );
        } 

        if ( thisboard->num_ports <= 0 || thisboard->num_ports > 5 ) {
                apci_error("num_ports is invalid: %d\n", thisboard->num_ports );
                return -EIO;
        }
        if ( thisboard->num_groups <= 0 || thisboard->num_groups > 5 ) {
                apci_error("num_groups is invalid: %d\n", thisboard->num_ports );
                return -EIO;
        }

	if (alloc_subdevices(dev, thisboard->num_ports*thisboard->num_groups) < 0)
		return -ENOMEM;
	
        apci_debug("Setting up device\n");
        retval = thisboard->setup_device( dev );
        if ( retval < 0 )
            return retval;

        if ( thisboard->irqhandler && pcidev->irq ) {
                /* TODO: eliminate this setup behavior */
                /* outb(0xF, devpriv->regions[2].start + 0x1 ); */

                if ((ret = comedi_request_irq( pcidev->irq,
                                              thisboard->irqhandler,
                                              IRQF_SHARED,
                                              driver_apci.driver_name,
                                              dev)) < 0 ) {
                        apci_error(" Irq %d not available\n", dev->irq );
                        dev->irq = 0;
                } else {
                        apci_info("Irq setup using %d\n",pcidev->irq );
                        dev->irq = pcidev->irq;
                        apci_debug("We are of type pci%s\n", thisboard->is_pcie ? "e" : "" );
                        devpriv->is_pcie = 1;
                }

        } else {
                apci_error("Has no IRQ handler\n");
        }


        if ( thisboard->initialize )
            thisboard->initialize( dev );

	apci_debug("attached\n");

	return 0;
}




/**
 * @todo Deactivate the IRQs
 */
static int apci_detach(comedi_device * dev)
{
        if (thisboard->teardown_device ) {
                apci_error("Tearing down device\n");
                thisboard->teardown_device( dev );
        }

        if (devpriv) {
                apci_error("Cleaning devpriv\n");
                if (devpriv->pcidev) {
                        if (devpriv->got_regions) {
                                comedi_pci_disable(devpriv->pcidev);
                        }
                        pci_dev_put(devpriv->pcidev);
                }
        }

        /* if (devpriv->plx_region_start ) { */
        /*         apci_debug("Releasing plx_region_start=%#x,len=%#x\n", devpriv->plx_region_start, devpriv->plx_region_length ); */
        /*         release_region( devpriv->plx_region_start, devpriv->plx_region_length); */
        /* } */
        /* cleanup_io_regions( dev ); */

        if (dev->irq ) {
                apci_info("removing the irq %d\n", dev->irq );
                comedi_free_irq(dev->irq, dev);
        }

        apci_debug("Detached\n");
	return 0;
}

int get_port_value( PORT port )
{
    switch (port ) {
    case PORT_A:
        return PORT_A;
    case PORT_B:
        return PORT_B;
    case PORT_C:
        return PORT_C;
    default:
        return PORT_A;
    }
}

static int teardown_device_iiro( comedi_device *dev )
{
        unsigned long flags;
        comedi_spin_lock_irqsave(&devpriv->lock,flags);
        outb( 0x0, devpriv->regions[2].start + 0x02 );
        comedi_spin_unlock_irqrestore(&devpriv->lock, flags);
                
        return 0;
}

static int teardown_device_idio( comedi_device *dev )
{
        comedi_subdevice *s;
        int i = 0;
        unsigned long flags;

        for ( i = 0 ; i < 3 ;i ++ ) {
                s = dev->subdevices + i;
                if ( s->private )  {
                        apci_error("Freeing private data\n");
                        kfree(s->private);
                }
        }

        comedi_spin_lock_irqsave(&devpriv->lock,flags);

        apci_debug("Disabling IDIO interrupts\n");
        outb( 0x0, devpriv->regions[2].start + 0xE );
        outb( 0x0, devpriv->regions[2].start + 0xF );
        outb( 0x0, devpriv->regions[2].start + 0x8 );
        outb( 0x0, devpriv->regions[2].start + 0x9 );
        outb( 0x0, devpriv->regions[2].start + 0xA );

        comedi_spin_unlock_irqrestore(&devpriv->lock, flags);

        return 0;
}

static int apci_iiro_initialize( comedi_device * dev )
{
        /* initialize the ports to all zero */
        apci_debug("Initializing %s \n", thisboard->name );
        outb(0x0, devpriv->regions[2].start + 0x0 );
        return 0;
}

static int apci_idio_initialize( comedi_device * dev )
{
        apci_debug("Initializing %s \n", thisboard->name );


        outb(0x0, devpriv->regions[2].start + 0x8 );
        outb(0x0, devpriv->regions[2].start + 0x9 );
        outb(0x0, devpriv->regions[2].start + 0xA );
        return 0;
}

static int apci_idio_cancel( comedi_device * dev, comedi_subdevice * s )
{
        if ( s == &dev->subdevices[1] ) { 
                apci_debug("Cancelling idio interrupt for CMOS TTL\n");                
        } else {
                apci_debug("Cancelling idio interrupt for ISOL\n");                
        }
        disable_idio_interrupt( dev, s );

        return 0;
}

static int apci_idio_ttl_cmdtest( comedi_device * dev, comedi_subdevice * s, comedi_cmd *cmd )
{
	int err = 0;
	int tmp;
	/* step 1: make sure trigger sources are trivially valid */
        apci_info("comedi%d: Calling cmd_test for subdev %s\n",dev->minor, ( s == &dev->subdevices[1] ? "CMOS" : "ISOL" ) );
        cmd->start_src = TRIG_INT;
        apci_info("Start src is %d\n", cmd->start_src );

	tmp = cmd->start_src;
	cmd->start_src &= (TRIG_NOW | TRIG_INT);
	if (!cmd->start_src || tmp != cmd->start_src) {
            apci_debug("Error cmd->start_src\n");
		err++;
        }

	tmp = cmd->scan_begin_src;
	cmd->scan_begin_src &= 	TRIG_EXT | TRIG_TIMER;
	if (!cmd->scan_begin_src || tmp != cmd->scan_begin_src) {
                apci_debug("Error cmd->scan_begin_src\n");
                err++;
        }

	tmp = cmd->convert_src;
	cmd->convert_src &= TRIG_NOW;
	if (!cmd->convert_src || tmp != cmd->convert_src) {
                apci_debug("Error cmd->convert_src\n");
		err++;
        }

	tmp = cmd->scan_end_src;
	cmd->scan_end_src &= TRIG_COUNT;

	if (!cmd->scan_end_src || tmp != cmd->scan_end_src) {
                apci_debug("Error cmd->scan_end_src\n");
		err++;
        }

	tmp = cmd->stop_src;

	cmd->stop_src &= TRIG_NONE | TRIG_COUNT ;
	if (!cmd->stop_src || tmp != cmd->stop_src) {
                apci_debug("Error cmd->stop_src\n");
		err++;
        }
	if (err)
		return 1;

	/* step 2: make sure trigger sources are unique and mutually compatible */

	if (cmd->convert_src != TRIG_NOW && cmd->convert_src != TRIG_TIMER)
		err++;
	if (cmd->stop_src != TRIG_COUNT && cmd->stop_src != TRIG_NONE)
		err++;

	if (err)
		return 2;

	/* step 3: make sure arguments are trivially compatible */

	if (cmd->start_arg != 0) {
		cmd->start_arg = 0;
		err++;
	}

	if (cmd->convert_src == TRIG_NOW) {
		if (cmd->convert_arg != 0) {
			cmd->convert_arg = 0;
			err++;
		}
	}
	if (cmd->scan_begin_src == TRIG_TIMER) {
		if (cmd->scan_begin_arg < nano_per_micro) {
			cmd->scan_begin_arg = nano_per_micro;
			err++;
		}
		if (cmd->convert_src == TRIG_TIMER &&
			cmd->scan_begin_arg <
			cmd->convert_arg * cmd->chanlist_len) {
			cmd->scan_begin_arg =
				cmd->convert_arg * cmd->chanlist_len;
			err++;
		}
	}
	// XXX these checks are generic and should go in core if not there already
	if (!cmd->chanlist_len) {
		cmd->chanlist_len = 1;
		err++;
	}
        
        apci_debug("USing scna_end_arg of %d, instead %d\n", cmd->scan_end_arg, cmd->chanlist_len );
	if (cmd->scan_end_arg != cmd->chanlist_len) {
		cmd->scan_end_arg = cmd->chanlist_len;
		err++;
	}

	if (cmd->stop_src == TRIG_COUNT) {
		if (!cmd->stop_arg) {
			cmd->stop_arg = 1;
			err++;
		}
	} else {		/* TRIG_NONE */
		if (cmd->stop_arg != 0) {
			cmd->stop_arg = 0;
			err++;
		}
	}

	if (err)
		return 3;

	/* step 4: fix up any arguments */

	if (cmd->scan_begin_src == TRIG_TIMER) {
		tmp = cmd->scan_begin_arg;
		// round to nearest microsec
		cmd->scan_begin_arg = nano_per_micro * ((tmp + (nano_per_micro / 2)) / nano_per_micro);
		if (tmp != cmd->scan_begin_arg)
			err++;
	}
	if (cmd->convert_src == TRIG_TIMER) {
		tmp = cmd->convert_arg;
		// round to nearest microsec
		cmd->convert_arg = nano_per_micro * ((tmp + (nano_per_micro / 2)) / nano_per_micro);
		if (tmp != cmd->convert_arg)
			err++;
	}

	if (err)
		return 4;

	return 0;

}

static int apci_idio_ttl_cmd( comedi_device * dev, comedi_subdevice *s ) 
{
        /* Fire off the irq for the CMOS */
	comedi_cmd *cmd = &s->async->cmd;

	if (cmd->flags & TRIG_RT) {
		comedi_error(dev,
                             "commands at RT priority not supported in this driver");
		return -1;
	}

        subdevpriv->numevents = 0;

        if ( cmd->start_src == TRIG_INT ) {
                apci_debug("Trig Int events\n");
                start_idio_interrupt( dev, s );
                comedi_event( dev, s);
        } else if ( cmd->start_src == TRIG_NOW ) { 
                apci_debug("Trig Now events\n");
                start_idio_interrupt( dev, s );
                comedi_event( dev, s);
        } else {
                comedi_error(dev, "bug setting conversion period");
                return -1;
        }
  
        switch (cmd->stop_src) {
        case TRIG_COUNT:
                subdevpriv->numevents = cmd->stop_arg;
                subdevpriv->continuous = 0;
                break;
        case TRIG_NONE:
                apci_error("Performing continuous acquisition\n");
                subdevpriv->continuous = 1;
                break;
        }

        return 0;
}

static int apci_iiro_intr_cmd(comedi_device * dev, comedi_subdevice * s)
{
	comedi_cmd *cmd = &s->async->cmd;
        int event;
        start_interrupt( dev );
	if (cmd->flags & TRIG_RT) {
		comedi_error(dev,
			"commands at RT priority not supported in this driver");
		return -1;
	}

        devpriv->numevents = 0;

        if ( cmd->start_src == TRIG_INT ) {
          apci_info("Doing some TRIG_INT\n");
          start_interrupt(dev);
        } else {
          apci_info("Starting the interrupt !\n");
          event = start_interrupt(dev );
          comedi_event( dev, s);
        }

	if (cmd->convert_src == TRIG_NOW)
		/* devpriv->convert_period = 0; */
                apci_info("comedi%d: Doing some TRIG_NOW stuff\n", dev->minor);
	else if (cmd->convert_src == TRIG_TIMER)
		/* devpriv->convert_period = cmd->convert_arg / nano_per_micro; */
                apci_info("comedi%d: Doing some TRIG_TIMER stuff\n",dev->minor);
	else {
                comedi_error(dev, "bug setting conversion period");
                return -1;
	}

        switch (cmd->stop_src) {
        case TRIG_COUNT:
                devpriv->numevents = cmd->stop_arg;
                devpriv->continuous = 0;
                break;
        case TRIG_NONE:
                apci_error("Performing continuous acquisition\n");
                devpriv->continuous = 1;
                devpriv->numevents = 1;
                break;
        }

        return 0;
}



static int apci_iiro_intr_cmdtest(comedi_device * dev, comedi_subdevice * s,
        comedi_cmd * cmd)
{
	int err = 0;
	int tmp;

	/* step 1: make sure trigger sources are trivially valid */
        apci_info("comedi%d: Calling cmd_test\n",dev->minor);
        apci_info("Start src is %d\n", cmd->start_src );
	tmp = cmd->start_src;
	cmd->start_src &= (TRIG_NOW | TRIG_INT);
	if (!cmd->start_src || tmp != cmd->start_src)
		err++;


	tmp = cmd->scan_begin_src;
	/* cmd->scan_begin_src &= ( TRIG_TIMER | TRIG_FOLLOW ); */
	cmd->scan_begin_src &= ( TRIG_TIMER | TRIG_FOLLOW );
	if (!cmd->scan_begin_src || tmp != cmd->scan_begin_src)
		err++;


	tmp = cmd->convert_src;
	/* cmd->convert_src &= TRIG_NOW | TRIG_TIMER; */
	cmd->convert_src &= TRIG_NOW;
	if (!cmd->convert_src || tmp != cmd->convert_src)
		err++;


	tmp = cmd->scan_end_src;
	cmd->scan_end_src &= TRIG_COUNT;
        apci_info("comedi_test: scan_end_src: %d\n", cmd->scan_end_src );
	if (!cmd->scan_end_src || tmp != cmd->scan_end_src)
		err++;

	tmp = cmd->stop_src;
        apci_info("comedi_test: stop_src: %d\n", cmd->stop_src );
	cmd->stop_src &= TRIG_NONE | TRIG_COUNT ;
	if (!cmd->stop_src || tmp != cmd->stop_src)
		err++;

	if (err)
		return 1;

	/* step 2: make sure trigger sources are unique and mutually compatible */

	if (cmd->convert_src != TRIG_NOW && cmd->convert_src != TRIG_TIMER)
		err++;
	if (cmd->stop_src != TRIG_COUNT && cmd->stop_src != TRIG_NONE)
		err++;

	if (err)
		return 2;

	/* step 3: make sure arguments are trivially compatible */

	if (cmd->start_arg != 0) {
		cmd->start_arg = 0;
		err++;
	}

	if (cmd->convert_src == TRIG_NOW) {
		if (cmd->convert_arg != 0) {
			cmd->convert_arg = 0;
			err++;
		}
	}
	if (cmd->scan_begin_src == TRIG_TIMER) {
		if (cmd->scan_begin_arg < nano_per_micro) {
			cmd->scan_begin_arg = nano_per_micro;
			err++;
		}
		if (cmd->convert_src == TRIG_TIMER &&
			cmd->scan_begin_arg <
			cmd->convert_arg * cmd->chanlist_len) {
			cmd->scan_begin_arg =
				cmd->convert_arg * cmd->chanlist_len;
			err++;
		}
	}
	// XXX these checks are generic and should go in core if not there already
	if (!cmd->chanlist_len) {
		cmd->chanlist_len = 1;
		err++;
	}
	if (cmd->scan_end_arg != cmd->chanlist_len) {
		cmd->scan_end_arg = cmd->chanlist_len;
		err++;
	}

	if (cmd->stop_src == TRIG_COUNT) {
		if (!cmd->stop_arg) {
			cmd->stop_arg = 1;
			err++;
		}
	} else {		/* TRIG_NONE */
		if (cmd->stop_arg != 0) {
			cmd->stop_arg = 0;
			err++;
		}
	}

	if (err)
		return 3;

	/* step 4: fix up any arguments */

	if (cmd->scan_begin_src == TRIG_TIMER) {
		tmp = cmd->scan_begin_arg;
		// round to nearest microsec
		cmd->scan_begin_arg =
			nano_per_micro * ((tmp +
				(nano_per_micro / 2)) / nano_per_micro);
		if (tmp != cmd->scan_begin_arg)
			err++;
	}
	if (cmd->convert_src == TRIG_TIMER) {
		tmp = cmd->convert_arg;
		// round to nearest microsec
		cmd->convert_arg =
			nano_per_micro * ((tmp +
				(nano_per_micro / 2)) / nano_per_micro);
		if (tmp != cmd->convert_arg)
			err++;
	}

	if (err)
		return 4;

	return 0;

}


static int apci_idio_insn_ttl_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data ) 
{
        /* subdev_private *spriv = s->private; */

        if (insn->n != 2)
		return -EINVAL;

	if (data[0]) {

            apci_debug("io_bits: %s\n", debug_byte( s->io_bits ));

            s->state &= ~data[0];
            s->state |= data[0] & data[1];

            apci_debug("writing '%s' to %#x\n", debug_byte(s->state), devpriv->regions[2].start + 0x3  );
            outb( s->state, devpriv->regions[2].start + 0x3  );
	}
        apci_debug("reading from %#x\n", devpriv->regions[2].start + 0x3 );
        data[1] = inb( devpriv->regions[2].start + 0x3 );
        apci_debug("read '%s'\n", debug_byte(data[1]) );

        return 2;
}

static int apci_idio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
        if (insn->n != 2)
		return -EINVAL;

        apci_debug("reading from %#x + (0x4,0x5,0x6)\n", devpriv->regions[2].start  );
        data[1]  = inb( devpriv->regions[2].start + 0x4 );
        data[1] |= inb( devpriv->regions[2].start + 0x5 ) << 8;
        data[1] |= inb( devpriv->regions[2].start + 0x6 ) << 16;
        apci_debug("read '%s'\n", debug_int(data[1]) );
        return 4;
}

static int apci_idio_fetout_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
        char a,b,c;
        if (insn->n != 2)
		return -EINVAL;

	if (data[0]) {

            /* apci_debug("io_bits: %s\n", debug_int( s->io_bits )); */
            apci_debug("state %s\n", debug_int( s->state ));
            s->state &= ~data[0];
            apci_debug("state %s\n", debug_int( s->state ));
            s->state |= (data[0] & data[1]);
            apci_debug("state %s\n", debug_int( s->state ));

            if (data[0] & 0xff) {
                    apci_debug("writing '%s' to %#x\n", debug_byte(s->state), devpriv->regions[2].start );
                    outb(s->state & 0xff, devpriv->regions[2].start + 0x0 );
            }
            if (data[0] & 0xff00) {
                    apci_debug("writing '%s' to %#x\n", debug_byte((s->state >> 8) & 0xff ), devpriv->regions[2].start + 1);
                    outb((s->state >> 8) & 0xff, devpriv->regions[2].start + 0x1 );
            }
            if (data[0] & 0xff0000 ) {
                    apci_debug("writing '%s' to %#x\n", debug_byte((s->state >> 16) & 0xff ), devpriv->regions[2].start + 2);
                    outb((s->state >> 16) & 0xff, devpriv->regions[2].start + 0x2 );
            }
	}
        a = inb(devpriv->regions[2].start + 0x0);
        b = inb(devpriv->regions[2].start + 0x1);
        c = inb(devpriv->regions[2].start + 0x2);

        apci_debug("addr=%#x   val=%s\n", devpriv->regions[2].start + 0x0, debug_byte(a));
        apci_debug("addr=%#x   val=%s\n", devpriv->regions[2].start + 0x1, debug_byte(b));
        apci_debug("addr=%#x   val=%s\n", devpriv->regions[2].start + 0x2, debug_byte(c));

        apci_debug("vals=(%d,%d,%d)\n",  c&0xff,b&0xff,a&0xff );

        data[1]  = a;
        data[1] |= (b << 8);
        data[1] |= (c << 16);

        apci_debug("idio read '%s'\n", debug_int( data[1] & 0x00ffffff ) );

        return 4;
}

static int apci_iiro_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
        subdev_private *spriv = s->private;

        if (insn->n != 2)
		return -EINVAL;

	if (data[0]) {

            /* apci_debug("io_bits: %s\n", debug_byte( s->io_bits )); */

            s->state &= ~data[0];
            s->state |= data[0] & data[1];

            apci_debug("writing '%s' to %#x\n", debug_byte(s->state), devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
            outb( s->state, devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
	}
        apci_debug("reading from %#x\n", devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
        data[1] = 1;
        data[1] = inb( devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
        apci_debug("read '%s'\n", debug_byte(data[1]) );

        return 2;
}

static int apci_dio_insn_bits(comedi_device * dev, comedi_subdevice * s,
                              comedi_insn * insn, lsampl_t * data)
{
        subdev_private *spriv = s->private;

	if (insn->n != 2)
		return -EINVAL;
	
	if (data[0]) {

            apci_debug("io_bits: %s\n",debug_byte( s->io_bits ));
            if ( (s->io_bits & data[0]) != data[0] ) 
                return -EIO;

            s->state &= ~data[0];
            s->state |= data[0] & data[1];

            apci_debug("writing '%s' to %#x\n", debug_byte(s->state), devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
            outb( s->state, devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
	}
        apci_debug("reading from %#x\n", devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
        data[1] = inb( devpriv->regions[2].start + spriv->group*4 + get_port_value(spriv->port) );
        apci_debug("read '%s'\n", debug_byte(data[1]) );

	return 2;
}

unsigned char calculate_direction( subdev_private *spriv, unsigned char omode, unsigned char bits, int channel )
{
    unsigned char retval;
    unsigned char incval;
    unsigned char exclval = -1;
    int invbits = ( bits == 0 ? 1 : 0 );

    switch ( spriv->port ) {
    case PORT_A:
        incval   = MAKE_BYTE(0,0,0,invbits,0,0,0,0);
        exclval  = MAKE_BYTE(1,1,1,0,1,1,1,1);
        break;
    case PORT_B:
        incval   = MAKE_BYTE(0,0,0,0,0,0,invbits,0);
        exclval  = MAKE_BYTE(1,1,1,1,1,1,0,1);
        break;
    case PORT_C:
        if ( channel <= 3 ) {
            incval   = MAKE_BYTE(0,0,0,0,0,0,0,(bits & 0xf)==0?1:0);
            exclval  = MAKE_BYTE(1,1,1,1,1,1,1,0);
        } else {
            incval   = MAKE_BYTE(0,0,0,0,(bits&0xf0)==0?1:0,0,0,0);
            exclval  = MAKE_BYTE(1,1,1,1,0,1,1,1);
        }
        break;
    default:
        incval = 0xff;
    }
    retval = (exclval & omode) | ( incval ); 

    return retval;
}

int make_io_bits( comedi_subdevice *s , int channel, int is_output ) 
{
    subdev_private *spriv = s->private;
    switch ( spriv->port ) {
    case PORT_A:
        if ( is_output ) 
            return 0xff;
        else
            return 0;
        break;
    case PORT_B:
        if ( is_output ) 
            return 0xff;
        else 
            return 0;
        break;
    case PORT_C:
        if ( is_output ) {
            if ( channel <= 3 ) {
                s->io_bits |= 0x0f;
            } else {
                s->io_bits |= 0xf0;
            }
        } else {
            if ( channel <= 3 ) { 
                s->io_bits &= 0xf0;
            } else {
                s->io_bits &= 0x0f;
            }
        }
        break;
    default:
        break;
    }
    apci_debug("make_io_bits: returning %s\n", debug_byte( s->io_bits ));
    return s->io_bits;
}

/* static int apci_idio_insn_config(comedi_device * dev, comedi_subdevice * s, */
/* 	comedi_insn * insn, lsampl_t * data) */
/* { */

/* 	int chan = CR_CHAN(insn->chanspec); */
/*         int is_output = 0; */
/*         unsigned long flags; */
/*         subdev_private *spriv; */
/*         spriv = s->private; */
/*         apci_debug("Configuring...using channel: %d\n", chan ); */
	
/* 	switch (data[0]) { */
/* 	case INSN_CONFIG_DIO_OUTPUT: */
/*                 /\* apci_debug("Setting all channels of Group %d Port %c to output ", spriv->group,debug_port( spriv->port) ); *\/ */
/*                 s->io_bits = make_io_bits( s, chan, 1 ); */
/*                 apci_debug("%s\n",debug_byte(s->io_bits)); */
/*                 is_output = 1; */
/*                 break; */
/* 	case INSN_CONFIG_DIO_INPUT: */
/*                 /\* apci_debug("Setting all channels of Group %d Port %c to input ",  spriv->group, debug_port( spriv->port ) ); *\/ */
/*                 s->io_bits = make_io_bits( s, chan, 0 ); */
/*                 apci_debug("%s\n",debug_byte(s->io_bits)); */
/*                 is_output = 0; */
/*                 break; */
/* 	case INSN_CONFIG_DIO_QUERY: */
/* 		data[1] = (s-> io_bits & (1 << chan)) ? COMEDI_OUTPUT : COMEDI_INPUT; */
/* 		return insn->n; */
/* 		break; */
/* 	default: */
/* 		return -EINVAL; */
/* 		break; */
/* 	} */

/*         spriv = s->private; */
/*         if ( spriv ) { */
/*                 apci_debug("Sending signal to Base + %d\n", (spriv->port + spriv->group*4 )); */
/*         } */

/*         comedi_spin_lock_irqsave(&devpriv->lock, flags); */
/*         apci_debug("Before: %s\n", debug_byte(devpriv->mode[spriv->group]) ); */
/*         devpriv->mode[spriv->group] = calculate_direction( spriv, devpriv->mode[spriv->group], s->io_bits , chan  ); */
/*         apci_debug("After:  %s\n", debug_byte(devpriv->mode[spriv->group]) ); */
/*         apci_debug("Sending write to + %#x\n", devpriv->regions[2].start + spriv->group*4 + 3 ); */
/*         outb(devpriv->mode[spriv->group] , devpriv->regions[2].start + spriv->group*4 + 3 ); */
/*         comedi_spin_unlock_irqrestore(&devpriv->lock, flags); */

/*         return -1; */
/* } */

static int apci_idio_insn_ttl_config( comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
	int chan = CR_CHAN(insn->chanspec);
        unsigned long flags;
        subdev_private *spriv;
        unsigned char tmp;

        spriv = s->private;

        comedi_spin_lock_irqsave(&devpriv->lock, flags);	
	switch (data[0]) {
	case INSN_CONFIG_DIO_OUTPUT:
                apci_debug("All bits of TTL to input \n");
                tmp = inb( devpriv->regions[2].start + 0xc ) | ( 2 );
                outb( tmp, devpriv->regions[2].start + 0xc );
                break;
	case INSN_CONFIG_DIO_INPUT:
                apci_debug("All bits of TTL to output\n");
                tmp = inb( devpriv->regions[2].start + 0xc ) & (~2);
                outb( tmp, devpriv->regions[2].start + 0xc );
                break;
	case INSN_CONFIG_DIO_QUERY:
		data[1] = (s-> io_bits & (1 << chan)) ? COMEDI_OUTPUT : COMEDI_INPUT;
		return insn->n;
		break;
	default:
		return -EINVAL;
		break;
	}
        comedi_spin_unlock_irqrestore(&devpriv->lock, flags);
        return insn->n;
}

static int apci_dio_insn_config(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
	int chan = CR_CHAN(insn->chanspec);
        int is_output = 0;
        unsigned long flags;
        subdev_private *spriv;
        spriv = s->private;
        apci_debug("Configuring...using channel: %d\n", chan );
	
	switch (data[0]) {
	case INSN_CONFIG_DIO_OUTPUT:
            apci_debug("Setting all channels of Group %d Port %c to output ", spriv->group,debug_port( spriv->port) );
            s->io_bits = make_io_bits( s, chan, 1 );
            apci_debug("%s\n",debug_byte(s->io_bits));
            is_output = 1;
            break;
	case INSN_CONFIG_DIO_INPUT:
            apci_debug("Setting all channels of Group %d Port %c to input ",  spriv->group, debug_port( spriv->port ) );
            s->io_bits = make_io_bits( s, chan, 0 );
            apci_debug("%s\n",debug_byte(s->io_bits));
            is_output = 0;
            break;
	case INSN_CONFIG_DIO_QUERY:
		data[1] = (s-> io_bits & (1 << chan)) ? COMEDI_OUTPUT : COMEDI_INPUT;
		return insn->n;
		break;
	default:
		return -EINVAL;
		break;
	}

        spriv = s->private;
        if ( spriv ) {
            apci_debug("Sending signal to Base + %d\n", (spriv->port + spriv->group*4 ));
        }

        comedi_spin_lock_irqsave(&devpriv->lock, flags);


        apci_debug("Before: %s\n", debug_byte(devpriv->mode[spriv->group]) );

        devpriv->mode[spriv->group] = calculate_direction( spriv, devpriv->mode[spriv->group], s->io_bits , chan  );
        apci_debug("After:  %s\n", debug_byte(devpriv->mode[spriv->group]) );
        apci_debug("Sending write to + %#x\n", devpriv->regions[2].start + spriv->group*4 + 3 );
        outb(devpriv->mode[spriv->group] , devpriv->regions[2].start + spriv->group*4 + 3 );

        comedi_spin_unlock_irqrestore(&devpriv->lock, flags);

	return insn->n;
}

static int apci_iiro_insn_config(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
	int chan = CR_CHAN(insn->chanspec);
        unsigned long flags;
        subdev_private *spriv;
        spriv = s->private;
        apci_debug("Configuring...using channel: %d\n", chan );
	
	switch (data[0]) {
	case INSN_CONFIG_FILTER:
                apci_debug("Enabling filtering on Group=%d,port=%c", spriv->group, debug_iiro_port( spriv->port));
                break;

	case INSN_CONFIG_DIO_QUERY:
		data[1] = (s-> io_bits & (1 << chan)) ? COMEDI_OUTPUT : COMEDI_INPUT;
		return insn->n;
		break;
	default:
		return -EINVAL;
		break;
	}

        spriv = s->private;
        if ( spriv ) {
            apci_debug("Sending signal to Base + %d\n", (spriv->port + spriv->group*4 ));
        }

        comedi_spin_lock_irqsave(&devpriv->lock, flags);

        apci_debug("Before: %s\n", debug_byte(devpriv->mode[spriv->group]) );
        devpriv->mode[spriv->group] = calculate_direction( spriv, devpriv->mode[spriv->group], s->io_bits , chan  );
        apci_debug("After:  %s\n", debug_byte(devpriv->mode[spriv->group]) );
        apci_debug("Sending write to + %#x\n", devpriv->regions[2].start + spriv->group*4 + 3 );
        outb(devpriv->mode[spriv->group] , devpriv->regions[2].start + spriv->group*4 + 3 );

        comedi_spin_unlock_irqrestore(&devpriv->lock, flags);

	return insn->n;
}

int setup_device_iiro8( comedi_device *dev )
{
        subdev_private *spriv;
        comedi_subdevice *s;
        int group = 0;
        int retval = 0;

        apci_debug("Setting up %s device\n", thisboard->name );
        for ( group = 0; group < thisboard->num_groups ; group ++ ) { 
                /*----------------------------   Output  ------------------------*/
                s = dev->subdevices + 0 + (group*2);
                s->type = COMEDI_SUBD_DIO;
                s->subdev_flags  = SDF_READABLE | SDF_WRITABLE;
                s->n_chan        = 8;
                s->maxdata       = 1;
                s->range_table   = &range_digital;
                s->insn_bits     = apci_iiro_insn_bits;
                s->insn_config   = apci_iiro_insn_config;
                spriv = (subdev_private *)kmalloc(sizeof(subdev_private), GFP_KERNEL );
                if ( !spriv ) {
                        apci_error("comedi%d: error! out of memory!\n", dev->minor);
                        return -ENOMEM;
                }
                spriv->group = group;
                spriv->port  = PORT_OUT;
                s->private = spriv;

                /*-------------------  Isolated Input / Change of State COS  ------------------*/
                s = dev->subdevices + (group*2) + 1;
                dev->read_subdev = s;
                s->type = COMEDI_SUBD_DI;
                s->subdev_flags  = SDF_READABLE | SDF_CMD_READ;
                s->n_chan        = 8;
                s->maxdata       = 1;
                s->range_table   = &range_digital;
                s->insn_bits     = apci_iiro_insn_bits;
                s->insn_config   = apci_iiro_insn_config;
                s->do_cmdtest    = apci_iiro_intr_cmdtest;
                s->do_cmd        = apci_iiro_intr_cmd;

                spriv = (subdev_private *)kmalloc(sizeof(subdev_private), GFP_KERNEL );
                if ( !spriv ) {
                        apci_error("comedi%d: error! out of memory!\n", dev->minor);
                        return -ENOMEM;
                }
                spriv->group = group;
                spriv->port  = PORT_IN;
                s->private = spriv;
        }
        
        return retval;
}

int setup_device_idio( comedi_device *dev )
{
        subdev_private *spriv;
        comedi_subdevice *s;

        int retval = 0;

        apci_debug("Setting up %s device\n", thisboard->name );
        apci_info("Setting up OUTPUT\n");
        /*----------------------------   Output ONLY ------------------------*/
        s = dev->subdevices + 0;
        s->type = COMEDI_SUBD_DO;
        s->subdev_flags  = SDF_WRITABLE;
        s->n_chan        = 24;
        s->maxdata       = 1;
        s->range_table   = &range_digital;
        s->insn_bits     = apci_idio_fetout_bits;

        s->state = 0;
        s->private = NULL;


        apci_info("Setting up TTL CMOS\n");
        /*--------------------------------    TTL CMOS   ------------------------------*/
        s = dev->subdevices + 1;
        s->type = COMEDI_SUBD_DIO;
        s->subdev_flags  = SDF_READABLE | SDF_WRITABLE | SDF_CMD_READ;
        s->n_chan        = 8;
        s->len_chanlist  = 8;
        s->maxdata       = 1;
        s->range_table   = &range_digital;
        s->insn_bits     = apci_idio_insn_ttl_bits;
        s->insn_config   = apci_idio_insn_ttl_config;
        s->do_cmdtest    = apci_idio_ttl_cmdtest;
        s->do_cmd        = apci_idio_ttl_cmd;
        s->cancel        = apci_idio_cancel;
        s->state = 0;
        s->private = NULL;
        spriv = (subdev_private *)kmalloc(sizeof(subdev_private), GFP_KERNEL );
        if ( !spriv ) {
                apci_error("comedi%d: error! out of memory!\n", dev->minor);
                return -ENOMEM;
        }
        s->private = spriv;
        subdevpriv->state = inb( devpriv->regions[2].start+ 0x7);
        subdevpriv->continuous = 0;

        apci_info("Setting up Isolated Input / COS\n");
        /*-------------------  Isolated Input / ISOL / Change of State COS  ------------------*/
        s = dev->subdevices + 2;
        s->type = COMEDI_SUBD_DI;
        s->subdev_flags  = SDF_READABLE | SDF_CMD_READ | SDF_LSAMPL ;
        s->n_chan        = 24;
        s->len_chanlist  = 24;
        s->maxdata       = 1;
        s->range_table   = &range_digital;
        s->insn_bits     = apci_idio_insn_bits;
        s->do_cmdtest    = apci_idio_ttl_cmdtest;
        s->do_cmd        = apci_idio_ttl_cmd;
        s->cancel        = apci_idio_cancel;

        s->private = NULL;

        spriv = (subdev_private *)kmalloc(sizeof(subdev_private), GFP_KERNEL );
        if ( !spriv ) {
                apci_error("comedi%d: error! out of memory!\n", dev->minor);
                return -ENOMEM;
        }
        s->private = spriv;
        subdevpriv->state = inb( devpriv->regions[2].start+ 0x4) | (inb( devpriv->regions[2].start+ 0x5) << 8) | (inb( devpriv->regions[2].start+ 0x5) << 16);
        subdevpriv->continuous = 0;

        apci_debug("Clearing all register values \n");
        /* outl( 0x0, devpriv->regions[2].start ); */
        outb(0x0, devpriv->regions[2].start + 0xF );
        


        return retval;
}

static int start_interrupt( comedi_device *dev )
{
        apci_devel("Starting interrupts\n");
  
        inb( devpriv->regions[2].start + 0x2 );
        outb( 0x9, devpriv->plx_region_start + 0x69 );

        return 0;
}

static int start_idio_interrupt( comedi_device *dev, comedi_subdevice *s )
{
        if ( s == &dev->subdevices[1] ) { 
                apci_devel("Starting IDIO interrupts for CMOS\n");
                outb( (inb(devpriv->regions[2].start + 0xE ) | (( 1<< 7) | (8))) , devpriv->regions[2].start + 0xE );
        } else {
                apci_devel("Starting IDIO interrupts  for ISOL\n");
                outb( (inb(devpriv->regions[2].start + 0xE ) | ( 0x77) ) , devpriv->regions[2].start + 0xE );
        }

        return 0;
}

static int disable_idio_interrupt( comedi_device *dev , comedi_subdevice *s )
{
        if ( s == &dev->subdevices[1] ) { 
                apci_debug("Disabling IDIO for CMOS\n");
                outb( inb( devpriv->regions[2].start + 0xE ) & ~( (1 << 7) | ( 8 ))  , devpriv->regions[2].start + 0xE );
        } else {
                apci_devel("Disabling IDIO for ISOL\n");
                outb( inb( devpriv->regions[2].start + 0xE ) & ( (1 << 7) | (8)), devpriv->regions[2].start + 0xE );
        }

        return 0;
}



static int disable_interrupts( comedi_device *dev )
{
        apci_devel("Disabling interrupts\n");
        outb(0x1, devpriv->regions[2].start + 0x2 );
        return  0;
}

irqreturn_t iiro_irqhandler( int irq, void *d PT_REGS_ARG ) 
{
        comedi_device *dev  = d;
        unsigned long flags;
        unsigned char byte;
        sampl_t tmp;
        comedi_subdevice *s = dev->read_subdev;

        if( !devpriv->is_pcie ) { 
            byte = inb( devpriv->plx_region_start + 0x4C);
      
            if ((byte & 4) == 0) {
                return IRQ_NONE; /* not me */
            }
        } else {                    /* PCIe */
            byte = inb(devpriv->plx_region_start + 0x69);
            
            if ((byte & 0x80 ) == 0) {
                return IRQ_NONE; /* not me */
            }
        }

        apci_debug("Handling IRQ\n");
        comedi_spin_lock_irqsave(&dev->spinlock, flags);
        outb(0x0, devpriv->regions[2].start + 0x2 );
        outb(0xF, devpriv->regions[2].start + 0x1 ); /* Clear interrupt */

        byte = inb( devpriv->regions[2].start + 0x1 );

        apci_devel("Value was %d\n", (int)byte );
        tmp = (0x00ff) & ( byte );
        inb( devpriv->regions[2].start + 0x2 );
        comedi_buf_put(s->async, tmp );
        comedi_spin_unlock_irqrestore(&dev->spinlock, flags);

        if ( !devpriv->continuous )
                devpriv->numevents --;    

        if ( devpriv->numevents <= 0 ) {
                disable_interrupts( dev );
                devpriv->numevents = 0;
                apci_devel("Completed receiving events\n");
                s->async->events |= COMEDI_CB_EOA;
        }

        s->async->events |= COMEDI_CB_BLOCK | COMEDI_CB_EOS;
        comedi_event(dev, s);


        return IRQ_HANDLED;
}

irqreturn_t idio_irqhandler( int irq, void *d PT_REGS_ARG ) 
{
        comedi_device *dev  = d;
        unsigned long flags;
        unsigned char cmos_input;
        int isol_input;

        int clear_irq;

        comedi_subdevice *s;
        
        apci_debug("Handling IRQ\n");
        comedi_spin_lock_irqsave(&dev->spinlock, flags);

        /* Temporarily broken...at first clear all interrupts */
        outl( 0x0, devpriv->regions[2].start + 0x8 ); /* Clear all interrupts */

        clear_irq   = inl( devpriv->regions[2].start + 0x8 );
        cmos_input  = inb( devpriv->regions[2].start + 0x7 ) & 0xff;
        isol_input  = inb(devpriv->regions[2].start + 0x4 ) | ( inb(devpriv->regions[2].start + 0x5 ) << 8 ) | ( inb(devpriv->regions[2].start + 0x6 ) << 16 );
        apci_debug("Got value cmos=%d ,  isol=%d\n", (int)(cmos_input & 0xff), isol_input );

        outl( clear_irq,  devpriv->regions[2].start + 0x8 );

        comedi_spin_unlock_irqrestore(&dev->spinlock, flags);

        s = &dev->subdevices[1];
        if ( (subdevpriv->state & 0xff) != cmos_input  ) {
                /* actually a change on cmos input */
                apci_debug("Changed CMOS :before=%u,  after=%u\n", (unsigned)(subdevpriv->state & 0xff ), (unsigned)cmos_input );
                if ( s->async ) {
                        if ( !subdevpriv->continuous )
                                subdevpriv->numevents --;    

                        subdevpriv->state = cmos_input;

                        if ( subdevpriv->numevents <= 0 ) {
                                apci_idio_cancel( dev, s );
                                subdevpriv->numevents = 0;
                                s->async->events |= COMEDI_CB_EOA;
                        }

                        comedi_buf_put(s->async, cmos_input ) ;
                        s->async->events |= COMEDI_CB_BLOCK | COMEDI_CB_EOS;                        
                        comedi_event(dev, s);
                }
        }

        s = &dev->subdevices[2];
        if ( (subdevpriv->state & 0x00ffffff ) != isol_input ) {
                /* actually a change on isol inputs */
                apci_debug("Changed ISOL :before=%d,  after=%d\n", subdevpriv->state , isol_input );
                if ( s->async ) { 

                        if ( !subdevpriv->continuous )
                                subdevpriv->numevents --;    

                        apci_debug("Continuous: %d, numevents: %d, sizeisol=%d\n", subdevpriv->continuous, subdevpriv->numevents , (int)sizeof(isol_input));
                        subdevpriv->state = isol_input;

                        if ( subdevpriv->numevents <= 0 ) {
                                apci_idio_cancel( dev, s );
                                subdevpriv->numevents = 0;
                                s->async->events |= COMEDI_CB_EOA;
                        }
                        apci_debug("ret1 %d\n", (isol_input & 0x0000ffff ) );
                        apci_debug("ret2 %d\n", ((isol_input >> 16) & 0x00ff )); 
                        comedi_buf_put(s->async, (isol_input & 0x0000ffff ) );
                        comedi_buf_put(s->async, ((isol_input >> 16) & 0x00ff )); 
                        s->async->events |= COMEDI_CB_BLOCK | COMEDI_CB_EOS;                        
                        comedi_event(dev, s);
                }
        }

        return IRQ_HANDLED;
}

int setup_device_dio48( comedi_device *dev )
{
        subdev_private *spriv;
        comedi_subdevice *s;
        int group;
        int retval = 0;
        apci_debug("Inside of the setup_device_dio48 function\n");
        apci_debug("Number of groups %d\n", thisboard->num_groups );
        for ( group = 0; group < thisboard->num_groups ; group ++ ) {
                /*--------------------------------  A  --------------------------------*/
                s = dev->subdevices + 0 + (group*3);
                s->type          = COMEDI_SUBD_DIO;
                s->subdev_flags  = SDF_READABLE | SDF_WRITABLE;
                s->n_chan        = 8;
                s->maxdata       = 1;
                s->range_table   = &range_digital;
                s->insn_bits     = apci_dio_insn_bits;
                s->insn_config   = apci_dio_insn_config;
                spriv = (subdev_private*)kmalloc(sizeof(subdev_private), GFP_KERNEL );
                if (!spriv) {
                        apci_error("comedi%d: error! out of memory!\n", dev->minor);
                        return -ENOMEM;
                }
                spriv->port  = PORT_A;
                spriv->group = group;
                s->private = spriv;

                /*--------------------------------  B  --------------------------------*/
                s = dev->subdevices + 1 + (group*3);
	
                s->type          = COMEDI_SUBD_DIO;
                s->subdev_flags  = SDF_READABLE | SDF_WRITABLE;
                s->n_chan        = 8;
                s->maxdata       = 1;
                s->range_table   = &range_digital;
                s->insn_bits     = apci_dio_insn_bits;
                s->insn_config   = apci_dio_insn_config;
                spriv = (subdev_private*)kmalloc(sizeof(subdev_private), GFP_KERNEL );
                if (!spriv) {
                        apci_error("comedi%d: error! out of memory!\n", dev->minor);
                        return -ENOMEM;
                }
                spriv->port  = PORT_B;
                spriv->group = group;
                s->private = spriv;
        
                /*--------------------------------  CHI & CLO-------------------------------*/
                s = dev->subdevices + 2 + (group*3) ;

                s->type = COMEDI_SUBD_DIO;
                s->subdev_flags = SDF_READABLE | SDF_WRITABLE | SDF_CMD_READ;
                s->n_chan = 8;
                s->maxdata = 1;
                s->range_table = &range_digital;
                s->insn_bits = apci_dio_insn_bits;
                s->insn_config = apci_dio_insn_config;
                spriv = (subdev_private*)kmalloc(sizeof(subdev_private), GFP_KERNEL );
                if (!spriv) {
                        apci_error("comedi%d: error! out of memory!\n", dev->minor);
                        return -ENOMEM;
                }
                spriv->port  = PORT_C;
                spriv->group = group;
                s->private = spriv;
        }
        return retval;
}


COMEDI_PCI_INITCLEANUP(driver_apci, apci_pci_table)

