#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/comedidev.h>

#include "apci_common.h"


#define AIO_USB_SIZE 0

#define AIO_USB_START_AI_CONV	0
#define AIO_USB_AI_READ		0


#define INPUT 1
#define OUTPUT 0
#define MODE_SEL_MODE1 1
#define MODE_SEL_MODE0 0

#define MODE_ACTIVE_AND_TRISTATE 1
#define SIZEOF_ADDRESS_SPACE  0x20
 
#define MODE1_HIGH 1
#define MODE1_LOW  0
#define MODE0_HIGH 0
#define MODE0_LOW  0

#define MAKE_BYTE(A,B,C,D,E,F,G,H) ( ( (A&1) << 7 ) | ( (B&1) << 6 ) | ( (C&1) << 5 ) | ( (D&1) << 4 ) | \
                                     ( (E&1) << 3 ) | ( (F&1) << 2 ) | ( (G&1) << 1 ) | ( (H&1) ) )

typedef struct aio_usb_board_struct {
    const char *name;
    int ai_chans;
    int ai_bits;
    int have_dio;
    int dio_bits;
    int dev_id;
} aio_usb_board;


typedef struct usb_aio_teststruct {
	struct slot_aio_teststruct *slot;
	struct usb_device *udev;
	struct usb_interface *interface;
	u16 vendor;
	u16 product;
	u16 device;
	u32 serial;
	struct {
		__u8 addr;
		size_t size;
	} message_pipe, command_write, command_read, write_stream, read_stream;
	struct kref kref;
	u16 analog_out_shadow[2];
	u8 digital_out_shadow;
} usb_aiotest_t;


typedef enum {
    PORT_A = 0,
    PORT_B = 1,
    PORT_C = 2,
    PORT_CHI = 2,
    PORT_CLO = 3
} PORT;

typedef struct {
    int start;
    int length;
    int flags;
    int enabled;
    void *mapped_address;
} io_region_t;

typedef struct {
    int offset;
    PORT port;
    int group;
    int plx_region_start;
    int plx_region_length;
    io_region_t regions[10];
    spinlock_t lock;
    struct pci_dev *pdev;
    unsigned char mode[5];
    int got_regions;
    struct pci_dev *pci_dev;
} aio_usb_private;


#define ACCCES_USB_VENDOR_ID 0x160f


#define PCI_DIO_96 0x8003
#define PCI_DIO_48 0x8040





static const aio_usb_board aio_usb_boards[] = {
    {
    name:	"usb_dio_96",
    dev_id:  PCI_DIO_96 ,
    have_dio:1
    },
    {
    name:	"usb_dio_48",
    dev_id: PCI_DIO_48,
    ai_chans:8,
    ai_bits:	16,
    have_dio:0
    },
};

/* static DEFINE_MUTEX(aiousb_mutex); */


static const struct usb_device_id aio_usb_table[] =  {
    {USB_DEVICE( ACCCES_USB_VENDOR_ID, 0x8003 )},
    {}
};

MODULE_DEVICE_TABLE( usb, aio_usb_table );

#define thisboard ((const aio_usb_board *)dev->board_ptr)
#define devpriv ((aio_usb_private *)dev->private)


static int aio_usb_dio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static int aio_usb_dio_insn_config(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data);
static void aio_usb_delete(struct kref *kref);

static int aio_usb_probe(struct usb_interface *interface,
        const struct usb_device_id *id);
static void aio_usb_disconnect(struct usb_interface *interface);

char debug_port( PORT port )
{
    static char retval;
    switch (port) {
    case PORT_A:
        retval = 'A';
        break;
    case PORT_B:
        retval = 'B';
        break;
    case PORT_C:
        retval = 'C';
        break;
    default:
        retval =  '_' ;
        break;
    }
    return retval;
}

char *debug_byte( unsigned char byte ) 
{
    static char buf[9];
    int i;
    for( i = 7 ; i >= 0 ; i -- ) { 
        /* aio_usb_debug("%d", (int)( byte & (1 << i ))); */
        buf[7-i] = ( byte & ( 1 << i) ? '1' : '0' );
    }
    buf[8] = 0;
    return buf;
}


void cleanup_io_regions( comedi_device *dev )
{
    int i;
    for ( i = 0; i < 6; i ++) {
        if (devpriv->regions[i].start == 0) {
            continue; /* invalid region */
        }
        if (devpriv->regions[i].flags & IORESOURCE_IO) {
            release_region(devpriv->regions[i].start, devpriv->regions[i].length);
            aio_usb_debug("Releasing bar=%d,region=%#x,len=%#x\n", i, devpriv->regions[i].start, devpriv->regions[i].length );
        } else {
            iounmap(devpriv->regions[i].mapped_address);
            aio_usb_debug("Unmapping bar=%d,region=%#x,mapped=%lx\n", i, devpriv->regions[i].start, (long int)devpriv->regions[i].mapped_address );
            release_mem_region(devpriv->regions[i].start, devpriv->regions[i].length);
            aio_usb_debug("Releasing mem bar=%d,region=%#x,len=%#x\n", i, devpriv->regions[i].start, devpriv->regions[i].length );
        }
    }
}

static void aio_usb_delete(struct kref *kref)
{
	/* usb_aiotest_t *dev = to_aio_teststruct_dev(kref); */
        usb_aiotest_t *dev = container_of( kref, usb_aiotest_t, kref );

	usb_put_dev(dev->udev);
	kfree(dev);
}

static struct usb_driver aio_usb_driver = {
#ifdef COMEDI_COMPAT_HAVE_USB_DRIVER_OWNER
	.owner = THIS_MODULE,
#endif
	.name = "accesusb",
	.probe = aio_usb_probe,
	.disconnect = aio_usb_disconnect,
	.id_table = aio_usb_table,
};


static int aio_usb_probe(struct usb_interface *interface,
	const struct usb_device_id *id) 
{
	int retval = -ENOMEM;
	usb_aiotest_t *dev = NULL;


	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	int i;
	/* u8 fw; */

        aio_usb_devel("Inside the silly probe function\n");
        aio_usb_error("DOING SOMETHING!!!\n");

	// allocate memory for our device state and initialize it
	dev = kzalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		dev_err(&interface->dev, "Out of memory");
		goto error;
	}
	kref_init(&dev->kref);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	// Check endpoints
	iface_desc = interface->cur_altsetting;

	if (iface_desc->desc.bNumEndpoints != 5) {
		dev_err(&interface->dev, "Wrong number of endpoints.");
		retval = -ENODEV;
		goto error;
	}

	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		int direction = -1;
		endpoint = &iface_desc->endpoint[i].desc;
		switch (i) {
		case 0:{
				direction = USB_DIR_IN;
				dev->message_pipe.addr =
					endpoint->bEndpointAddress;
				dev->message_pipe.size =
					le16_to_cpu(endpoint->wMaxPacketSize);
			}
			break;
		case 1:{
				direction = USB_DIR_OUT;
				dev->command_write.addr =
					endpoint->bEndpointAddress;
				dev->command_write.size =
					le16_to_cpu(endpoint->wMaxPacketSize);
			}
			break;
		case 2:{
				direction = USB_DIR_IN;
				dev->command_read.addr =
					endpoint->bEndpointAddress;
				dev->command_read.size =
					le16_to_cpu(endpoint->wMaxPacketSize);
			}
			break;
		case 3:{
				direction = USB_DIR_OUT;
				dev->write_stream.addr =
					endpoint->bEndpointAddress;
				dev->write_stream.size =
					le16_to_cpu(endpoint->wMaxPacketSize);
			}
			break;
		case 4:{
				direction = USB_DIR_IN;
				dev->read_stream.addr =
					endpoint->bEndpointAddress;
				dev->read_stream.size =
					le16_to_cpu(endpoint->wMaxPacketSize);
			}
			break;
		}
		if ((endpoint->bEndpointAddress & USB_DIR_IN) != direction) {
			dev_err(&interface->dev, "Endpoint has wrong direction.");
			retval = -ENODEV;
			goto error;
		}
	}


	dev->vendor = le16_to_cpu(dev->vendor);
	dev->product = le16_to_cpu(dev->product);
	dev->device = le16_to_cpu(dev->device);
	dev->serial = le32_to_cpu(dev->serial);

	dev->digital_out_shadow = 0;

	// save our data pointer in this interface device a
	usb_set_intfdata(interface, dev);

	// let the user know what node this device is now attached to
	dev_info(&interface->dev, "USB DT9812 (%4.4x.%4.4x.%4.4x) #0x%8.8x\n",
		dev->vendor, dev->product, dev->device, dev->serial);


	return 0;

      error:
	if (dev) {
		kref_put(&dev->kref, aio_usb_delete );
	}
	return retval;
}

static void aio_usb_disconnect(struct usb_interface *interface)
{
        usb_aiotest_t *dev;
        /* usb_aio_usb_t *dev; */
        /* usb_aio_device *dev; */
        int minor = interface->minor;

        /* mutex_lock(&aiousb_mutex); */
        dev = usb_get_intfdata(interface);
        /* if (dev->slot) { */
        /*         /\* mutex_lock(&dev->slot->mutex); *\/ */
        /*         dev->slot->usb = NULL; */
        /*         /\* mutex_unlock(&dev->slot->mutex); *\/ */
        /*         dev->slot = NULL; */
        /* } */
        usb_set_intfdata(interface, NULL);
        /* mutex_unlock(&aiousb_mutex); */


        kref_put(&dev->kref, aio_usb_delete);

        dev_info(&interface->dev, "USB Aio_Usb #%d now disconnected\n", minor);
}

static int aio_usb_attach(comedi_device * dev, comedi_devconfig * it)
{

	return 0;
}


static int aio_usb_detach(comedi_device * dev)
{
        /* if (devpriv) { */
        /*     if (devpriv->pdev) { */
        /*         if (devpriv->got_regions) { */
        /*             /\* comedi_pci_disable(devpriv->pci_dev); *\/ */
        /*         } */
        /*         /\* pci_dev_put(devpriv->pdev); *\/ */
        /*     } */
        /* } */

        /* cleanup_io_regions( dev ); */
        int n;
        if (dev->subdevices) {
            for ( n = 0; n < dev->n_subdevices; n++) {
                comedi_subdevice *s = &dev->subdevices[n];
                aio_usb_private *spriv = s->private;
                if (spriv) {
                    aio_usb_info("freeing private for subdevice %d\n", n);
                    kfree(spriv);
                }
            }
        }
        aio_usb_debug("Detached\n");
	return 0;
}

static int aio_usb_dio_insn_bits(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
        /* aio_usb_private *spriv = s->private; */
	return 2;
}

static int aio_usb_dio_insn_config(comedi_device * dev, comedi_subdevice * s,
	comedi_insn * insn, lsampl_t * data)
{
	return insn->n;
}

static comedi_driver aio_usb_comedi_driver = {
	.module = THIS_MODULE,
	.driver_name = "accesusb",
	.attach = aio_usb_attach,
	.detach = aio_usb_detach,
};


static int __init usb_aio_usb_init(void)
{
        int result;

        // register with the USB subsystem 
        result = usb_register(&aio_usb_driver);
        if (result) {
                printk(KERN_ERR KBUILD_MODNAME
                        ": usb_register failed. Error number %d", result);
        }
        // register with comedi
        result = comedi_driver_register(&aio_usb_comedi_driver);
        if (result) {
                usb_deregister(&aio_usb_driver);
                printk(KERN_ERR KBUILD_MODNAME ": comedi_driver_register failed. Error number %d", result);
        }

        return result;
}

static void __exit usb_aio_usb_exit(void)
{
        // unregister with comedi
        comedi_driver_unregister(&aio_usb_comedi_driver);

        /* deregister this driver with the USB subsystem */
        usb_deregister(&aio_usb_driver);

}





module_init(usb_aio_usb_init);
module_exit(usb_aio_usb_exit);

MODULE_AUTHOR("Jimi Damon <jdamon@accesio.com>");
MODULE_DESCRIPTION("Comedi USB-DIO-96 driver");
MODULE_LICENSE("GPL");


#endif
