/**
 * Example of using commands - asynchronous input
 * Part of Comedilib
 * 
 * Example for testing the comedi_test module
 */

#include <stdio.h>
#include <comedilib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <cstring>
#include <stdlib.h>
#include "gtest/gtest.h"
#include "tap.h"

extern comedi_t *device;

struct parsed_options
{
    char *filename;
    double value;
    int subdevice;
    int subdevice2;
    int channel;
    int aref;
    int range;
    int verbose;
    int n_chan;
    int n_scan;
    double freq;
    int dosomething;
};


#define N_CHANS 256
static unsigned int chanlist[N_CHANS];
static comedi_range * range_info[N_CHANS];
static lsampl_t maxdata[N_CHANS];


int prepare_cmd_lib(comedi_t *dev, int subdevice, int n_scan, int n_chan, unsigned period_nanosec, comedi_cmd *cmd);

void do_cmd(comedi_t *dev,comedi_cmd *cmd);

void print_datum(lsampl_t raw, int channel_index);

void dump_cmd(FILE *out,comedi_cmd *cmd);
char *cmd_src(int src,char *buf);


char *cmdtest_messages[]={
    (char *)"success",
    (char *)"invalid source",
    (char *)"source conflict",
    (char *)"invalid argument",
    (char *)"argument conflict",
    (char *)"invalid chanlist",
};
struct parsed_options options;
int subdev_flags;
comedi_t *dev;
comedi_cmd c,*cmd=&c;
lsampl_t raw;
#define BUFSZ 1024
sampl_t buf[BUFSZ];

/**
 * 
 */
TEST(ComediTest,AnalogReadings ) {
    lsampl_t data;
    int ret;
    int subdev = 0;        
    int chan = 0;          
    int range = 0;         
    int aref = AREF_GROUND;
    int num_scans = options.n_scan;
    int num_channels = options.n_chan;
    int readscans;
    std::string prop("/sys/class/comedi/comedi0/read_buffer_pos");

    dev = comedi_open(options.filename);
    ASSERT_TRUE( dev );
    subdev_flags = comedi_get_subdevice_flags(dev, options.subdevice);
    
    RecordProperty("NumScans", num_scans);
    #if 0
    fprintf(stderr, "command before testing:\n");
    dump_cmd(stderr, cmd);
    #endif

    prepare_cmd_lib(dev, options.subdevice, num_scans, num_channels, 1e9 / options.freq, cmd);
    #if 0
    fprintf(stderr, "command before testing:\n");
    dump_cmd(stderr, cmd);
    #endif

    /**
     * Setup the Command structure
     */
    cmd->chanlist_len = 2;
    cmd->stop_arg = num_scans;

    /* verify no problems */
    ret = comedi_command_test(dev, cmd);
    ASSERT_GE( ret, 0 ) << "Comedi Testing of command for device was ok";
    
    /* make sure our scan number hasn't changed */
    ASSERT_EQ( cmd->stop_arg , num_scans );
    ASSERT_EQ( cmd->chanlist_len , 2 );    

    ret = comedi_command(dev, cmd);
    sleep(1);

    ASSERT_GE( ret, 0 ) << "Able to Submit comedi command\n";


    /* Verify that the buffer position has in fact moved
       to say num_scans scans remain in it */
    FILE *fp = fopen(prop.c_str(),"r");
    ASSERT_TRUE( fp ) << "Able to open sysfs property file";
    fscanf(fp,"%d",&readscans );
    
    ASSERT_EQ( readscans, 4*num_scans ) << "Able to read the fifo position 4*" << num_scans << "\n";
    
    fclose(fp);

    comedi_close( dev );
}

TEST(ComediSysfs, CanReadMaxReadBuffer ) {
    lsampl_t data;
    int ret;
    int subdev = 0;        
    int chan = 0;          
    int range = 0;         
    int aref = AREF_GROUND;
    FILE *fp;
    char buf[100];
    int bufsize;

    std::string sysfs_entry("/sys/class/comedi/comedi0/max_read_buffer_kb");
    fp = fopen(sysfs_entry.c_str(),"r");
    ASSERT_TRUE( fp ) << "Able to open file " << sysfs_entry << "\n";
    fscanf(fp,"%d", &bufsize );
    EXPECT_GE( bufsize, 0 );
    comedi_close( dev );
}
                                         
int main(int argc, char *argv[])
{
	int ret;
	int total=0;
	int i;
        cmd =&c;
        /* init_parsed_options( &options ); */

	memset(&options, 0, sizeof(options));
	/* The following variables used in this demo
	 * can be modified by command line
	 * options.  When modifying this demo, you may want to
	 * change them here. */
        options.filename = ( !getenv("COMEDI_DEVICE") ? (char *)"/dev/comedi0" : getenv("COMEDI_DEVICE"));

	options.subdevice = 0;
	options.subdevice2 = 1;
	options.channel = 1;
	options.range = 0;
	options.aref = AREF_GROUND;
	options.n_chan = 4;
	options.n_scan = 1000;
	options.freq = 5000.0;
        /* parse_options( &options, argc, argv ); */
	/* open the device */
	/* dev = comedi_open(options.filename); */
        ::testing::GTEST_FLAG(print_time) = false;

        testing::InitGoogleTest(&argc, argv);
        testing::TestEventListeners & listeners = testing::UnitTest::GetInstance()->listeners();

        return RUN_ALL_TESTS();  



	return 0;
}

/*
 * This prepares a command in a pretty generic way.  We ask the
 * library to create a stock command that supports periodic
 * sampling of data, then modify the parts we want. */
int prepare_cmd_lib(comedi_t *dev, int subdevice, int n_scan, int n_chan, unsigned scan_period_nanosec, comedi_cmd *cmd)
{
	int ret;

	memset(cmd,0,sizeof(*cmd));

	/* This comedilib function will get us a generic timed
	 * command for a particular board.  If it returns -1,
	 * that's bad. */
	ret = comedi_get_cmd_generic_timed(dev, subdevice, cmd, n_chan, scan_period_nanosec);
	if (ret<0){
		printf("comedi_get_cmd_generic_timed failed\n");
		return ret;
	}

	/* Modify parts of the command */
	cmd->chanlist = chanlist;
	cmd->chanlist_len = 1;
	if (cmd->stop_src == TRIG_COUNT) cmd->stop_arg = n_scan;

	return 0;
}

void print_datum(lsampl_t raw, int channel_index) {
	double physical_value;
	physical_value = comedi_to_phys(raw, range_info[channel_index], maxdata[channel_index]);
	printf("%#8.6g ",physical_value);
}


char *cmd_src(int src,char *buf)
{
	buf[0]=0;

	if(src&TRIG_NONE)strcat(buf,"none|");
	if(src&TRIG_NOW)strcat(buf,"now|");
	if(src&TRIG_FOLLOW)strcat(buf, "follow|");
	if(src&TRIG_TIME)strcat(buf, "time|");
	if(src&TRIG_TIMER)strcat(buf, "timer|");
	if(src&TRIG_COUNT)strcat(buf, "count|");
	if(src&TRIG_EXT)strcat(buf, "ext|");
	if(src&TRIG_INT)strcat(buf, "int|");
#ifdef TRIG_OTHER
	if(src&TRIG_OTHER)strcat(buf, "other|");
#endif

	if(strlen(buf)==0){
		sprintf(buf,"unknown(0x%08x)",src);
	}else{
		buf[strlen(buf)-1]=0;
	}

	return buf;
}


void dump_cmd(FILE *out,comedi_cmd *cmd)

{
        char buf[100];

        fprintf(out,"subdevice:      %d\n",
                cmd->subdev);

        fprintf(out,"start:      %-8s %d\n",
                cmd_src(cmd->start_src,buf),
                cmd->start_arg);

        fprintf(out,"scan_begin: %-8s %d\n",
                cmd_src(cmd->scan_begin_src,buf),
                cmd->scan_begin_arg);

        fprintf(out,"convert:    %-8s %d\n",
                cmd_src(cmd->convert_src,buf),
                cmd->convert_arg);

        fprintf(out,"scan_end:   %-8s %d\n",
                cmd_src(cmd->scan_end_src,buf),
                cmd->scan_end_arg);

        fprintf(out,"stop:       %-8s %d\n",
                cmd_src(cmd->stop_src,buf),
                cmd->stop_arg);
}

